//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/InstrumentItems.cpp
//! @brief     Implement class InstrumentItem and all its children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/InstrumentItems.h"
#include "Base/Const/Units.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem1D.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "GUI/Model/CatDevice/InstrumentItemCatalog.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/BackgroundItems.h"
#include "GUI/Model/Device/BeamWavelengthItem.h"
#include "GUI/Model/Device/DetectorItems.h"
#include "GUI/Model/Device/Instrument.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Model/Device/RectangularDetectorItem.h"
#include "GUI/Model/Device/SphericalDetectorItem.h"
#include "GUI/Support/XML/Backup.h"
#include "GUI/Util/CoordName.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/String.h"
#include "Sim/Simulation/DepthprobeSimulation.h"
#include "Sim/Simulation/OffspecSimulation.h"
#include "Sim/Simulation/ScatteringSimulation.h"

namespace {
namespace Tag {

const QString Id("Id");
const QString Name("Name");
const QString Description("Description");
const QString WithPolarizerAnalyzer("WithPolarizerAnalyzer");
const QString AnalyzerEfficiency("AnalyzerEfficiency");
const QString AnalyzerTotalTransmission("AnalyzerTotalTransmission");
const QString Beam("Beam");
const QString Polarization("Polarization");
const QString AnalyzerDirection("AnalyzerDirection");
const QString Background("Background");
const QString Detector("Detector");
const QString ZAxis("ZAxis");
const QString AlphaAxis("AlphaAxis");
const QString BaseData("BaseData");

} // namespace Tag

void setBeamDistribution(ParameterDistribution::WhichParameter which,
                         const BeamDistributionItem& item, ISimulation& simulation)
{
    auto P_par_distr = item.getParameterDistributionForName(which);
    if (P_par_distr)
        simulation.addParameterDistribution(*P_par_distr);
}

} // namespace

//  ************************************************************************************************
//  class InstrumentItem
//  ************************************************************************************************

InstrumentItem::InstrumentItem()
    : m_withPolarizerAnalyzer(false)
{
    m_id = QUuid::createUuid().toString();
    m_analyzerEfficiency.init("Analyzer efficiency", "Efficiency of the polarization analysis", 0.0,
                              Unit::unitless, 4 /* decimals */, RealLimits::limitless(),
                              "efficiency");
    m_analyzerTotalTransmission.init(
        "Analyzer transmission", "Total transmission of the polarization analysis", 1.0,
        Unit::unitless, 4 /* decimals */, RealLimits::nonnegative(), "transmission");
    m_polarization.init("Polarization (Bloch vector)",
                        "Polarization of the beam, given as the Bloch vector", Unit::unitless,
                        "polarization");
    m_analyzerDirection.init("Analyzer direction", "GisasDirection of the polarization analysis",
                             Unit::unitless, "analyzerDirection");
    m_background.init("Background", "");
}

InstrumentItem* InstrumentItem::createItemCopy() const
{
    const auto type = InstrumentItemCatalog::type(this);
    auto* copy = InstrumentItemCatalog::create(type);
    GUI::Util::copyContents(this, copy);
    return copy;
}

BeamItem* InstrumentItem::beamItem() const
{
    return m_beamItem.get();
}

bool InstrumentItem::alignedWith(const RealItem* item) const
{
    return shape() == item->shape();
}

bool InstrumentItem::withPolarizerAnalyzer() const
{
    return m_withPolarizerAnalyzer;
}

void InstrumentItem::setWithPolarizerAnalyzer(bool with)
{
    m_withPolarizerAnalyzer = with;
}

template <typename T>
T* InstrumentItem::beamItemCast() const
{
    return dynamic_cast<T*>(m_beamItem.get());
}

void InstrumentItem::writeTo(QXmlStreamWriter* w) const
{
    ASSERT(m_beamItem);

    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // id
    w->writeStartElement(Tag::Id);
    XML::writeAttribute(w, XML::Attrib::value, m_id);
    w->writeEndElement();

    // name
    w->writeStartElement(Tag::Name);
    XML::writeAttribute(w, XML::Attrib::value, m_name);
    w->writeEndElement();

    // description
    w->writeStartElement(Tag::Description);
    XML::writeAttribute(w, XML::Attrib::value, m_description);
    w->writeEndElement();

    // with polarizer or analyzer?
    w->writeStartElement(Tag::WithPolarizerAnalyzer);
    XML::writeAttribute(w, XML::Attrib::value, m_withPolarizerAnalyzer);
    w->writeEndElement();

    // polarization
    w->writeStartElement(Tag::Polarization);
    m_polarization.writeTo(w);
    w->writeEndElement();

    // analyzer direction
    w->writeStartElement(Tag::AnalyzerDirection);
    m_analyzerDirection.writeTo(w);
    w->writeEndElement();

    // analyzer efficiency
    w->writeStartElement(Tag::AnalyzerEfficiency);
    m_analyzerEfficiency.writeTo(w);
    w->writeEndElement();

    // analyzer transmission
    w->writeStartElement(Tag::AnalyzerTotalTransmission);
    m_analyzerTotalTransmission.writeTo(w);
    w->writeEndElement();

    // background
    w->writeStartElement(Tag::Background);
    m_background.writeTo(w);
    w->writeEndElement();

    // beam
    w->writeStartElement(Tag::Beam);
    m_beamItem->writeTo(w);
    w->writeEndElement();
}

void InstrumentItem::readFrom(QXmlStreamReader* r)
{
    ASSERT(m_beamItem);

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // id
        if (tag == Tag::Id) {
            XML::readAttribute(r, XML::Attrib::value, &m_id);
            XML::gotoEndElementOfTag(r, tag);

            // name
        } else if (tag == Tag::Name) {
            XML::readAttribute(r, XML::Attrib::value, &m_name);
            XML::gotoEndElementOfTag(r, tag);

            // description
        } else if (tag == Tag::Description) {
            XML::readAttribute(r, XML::Attrib::value, &m_description);
            XML::gotoEndElementOfTag(r, tag);

            // with polarizer or analyzer?
        } else if (tag == Tag::WithPolarizerAnalyzer) {
            XML::readAttribute(r, XML::Attrib::value, &m_withPolarizerAnalyzer);
            XML::gotoEndElementOfTag(r, tag);

            // polarization
        } else if (tag == Tag::Polarization) {
            m_polarization.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // analyzer direction
        } else if (tag == Tag::AnalyzerDirection) {
            m_analyzerDirection.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // analyzer efficiency
        } else if (tag == Tag::AnalyzerEfficiency) {
            m_analyzerEfficiency.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // analyzer transmission
        } else if (tag == Tag::AnalyzerTotalTransmission) {
            m_analyzerTotalTransmission.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // background
        } else if (tag == Tag::Background) {
            m_background.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // beam
        } else if (tag == Tag::Beam) {
            m_beamItem->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

//  ************************************************************************************************
//  class SpecularInstrumentItem
//  ************************************************************************************************

SpecularInstrumentItem::SpecularInstrumentItem()
{
    m_beamItem.reset(new SpecularBeamItem(this));
}

SpecularBeamItem* SpecularInstrumentItem::beamItem() const
{
    return beamItemCast<SpecularBeamItem>();
}

std::vector<int> SpecularInstrumentItem::shape() const
{
    return {beamItem()->inclinationAxisItem()->binCount()};
}

void SpecularInstrumentItem::updateToRealData(const RealItem* item)
{
    if (shape().size() != item->shape().size())
        throw Error("Error in SpecularInstrumentItem::updateToRealData: The type "
                    "of instrument is incompatible with passed data shape.");

    const auto& data = item->nativeDatafield()->axis(0);
    beamItem()->updateToData(data, item->nativeDataUnits());
}

bool SpecularInstrumentItem::alignedWith(const RealItem* item) const
{
    const QString native_units = item->nativeDataUnits();
    if (native_units == "nbins")
        return beamItem()->inclinationAngleItem()->uniformAlphaAxisSelected()
               && shape() == item->shape();

    if (!beamItem()->inclinationAngleItem()->pointwiseAlphaAxisSelected())
        return false;

    const auto* axisItem =
        dynamic_cast<const PointwiseAxisItem*>(beamItem()->inclinationAngleItem()->alphaAxisItem());
    ASSERT(axisItem);

    if (axisItem->nativeAxisUnits() != native_units)
        return false;

    const auto* instrumentAxis = axisItem->axis();
    if (!instrumentAxis)
        return false;

    if (!item->hasNativeData())
        return false;

    // TODO do not compare native axis with main data axis due to different units
    // remove native data from everywhere
    // https://jugit.fz-juelich.de/mlz/bornagain/-/issues/331
    //
    //    const auto& native_axis = item->nativeDatafield()->axis(0);
    //    return *instrumentAxis == native_axis;
    return true;
}

const ICoordSystem* SpecularInstrumentItem::createCoordSystem() const
{
    std::shared_ptr<Beam> beam = beamItem()->createBeam();
    auto* axis_item = beamItem()->inclinationAxisItem();
    if (auto* pointwise_axis = dynamic_cast<PointwiseAxisItem*>(axis_item)) {
        if (!pointwise_axis->axis()) // workaround for loading project
            return nullptr;
        Coords native_units =
            GUI::Util::CoordName::coordFromName(pointwise_axis->nativeAxisUnits());
        return new AngularReflectometryCoords(beam->wavelength(), *pointwise_axis->axis(),
                                              native_units);
    }

    return new AngularReflectometryCoords(beam->wavelength(), *axis_item->createAxis(1.0),
                                          Coords::DEGREES);
}

QString SpecularInstrumentItem::instrumentType() const
{
    return "Specular";
}

QString SpecularInstrumentItem::defaultName() const
{
    return "Specular";
}

//  ************************************************************************************************
//  class DepthprobeInstrumentItem
//  ************************************************************************************************

DepthprobeInstrumentItem::DepthprobeInstrumentItem()
{
    m_beamItem.reset(new SpecularBeamItem(this));

    auto* axisItem = beamItem()->inclinationAxisItem();
    axisItem->setMin(0.0);
    axisItem->setMax(1.0);
    axisItem->setBinCount(500);

    m_zAxis.initMin("Min", "Starting value below sample horizon", -100.0, Unit::nanometer,
                    RealLimits::limitless());
    m_zAxis.initMax("Max", "Ending value above sample horizon", 100.0, Unit::nanometer,
                    RealLimits::limitless());
}

void DepthprobeInstrumentItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    InstrumentItem::writeTo(w);
    w->writeEndElement();

    // z axis
    w->writeStartElement(Tag::ZAxis);
    m_zAxis.writeTo(w);
    w->writeEndElement();
}

void DepthprobeInstrumentItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            InstrumentItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // z axis
        } else if (tag == Tag::ZAxis) {
            m_zAxis.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

SpecularBeamItem* DepthprobeInstrumentItem::beamItem() const
{
    return beamItemCast<SpecularBeamItem>();
}

std::vector<int> DepthprobeInstrumentItem::shape() const
{
    return {}; // no certain shape to avoid linking to real data
}

void DepthprobeInstrumentItem::updateToRealData(const RealItem*)
{
    throw std::runtime_error("DepthprobeInstrumentItem::updateToRealData()");
}

QString DepthprobeInstrumentItem::defaultName() const
{
    return "Depthprobe";
}

QString DepthprobeInstrumentItem::instrumentType() const
{
    return "Depth probe";
}

DepthprobeSimulation* DepthprobeInstrumentItem::createSimulation(const MultiLayer& sample) const
{
    auto* simulation = new DepthprobeSimulation(sample);

    auto* const axis_item = beamItem()->inclinationAxisItem();
    auto axis = axis_item->createAxis(Units::deg);

    simulation->setBeamParameters(beamItem()->wavelength(), static_cast<int>(axis->size()),
                                  axis->min(), axis->max());

    simulation->setzSpan(m_zAxis.nbins(), m_zAxis.min(), m_zAxis.max());

    setBeamDistribution(ParameterDistribution::BeamWavelength, *beamItem()->wavelengthItem(),
                        *simulation);

    setBeamDistribution(ParameterDistribution::BeamInclinationAngle,
                        *beamItem()->inclinationAngleItem(), *simulation);
    m_coosys.reset(simulation->simCoordSystem());
    return simulation;
}

const ICoordSystem* DepthprobeInstrumentItem::createCoordSystem() const
{
    if (!m_coosys)
        throw std::runtime_error(
            "DepthprobeSimulation::createCoordSystem: DepthprobeSimulation not yet initialized");
    // TODO set pol filters ?
    return m_coosys.get(); // TODO NOW RESTORE createSimulation()->createCoordSystem();
}

//  ************************************************************************************************
//  class Instrument2DItem
//  ************************************************************************************************

Instrument2DItem::Instrument2DItem()
{
    m_beamItem.reset(new GISASBeamItem());
    m_detector.init("Detector", "");
}

void Instrument2DItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    InstrumentItem::writeTo(w);
    w->writeEndElement();

    // detector
    w->writeStartElement(Tag::Detector);
    m_detector.writeTo(w);
    w->writeEndElement();
}

void Instrument2DItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            InstrumentItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // detector
        } else if (tag == Tag::Detector) {
            m_detector.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

void Instrument2DItem::importMasks(const MaskContainerItem* maskContainer)
{
    detectorItem()->importMasks(maskContainer);
}

std::shared_ptr<Instrument> Instrument2DItem::createInstrument() const
{
    auto beam = beamItem()->createBeam();
    auto detector = detectorItem()->createDetector();
    detector->setDetectorNormal(beam->ki());

    return std::make_shared<Instrument>(beam, *detector);
}

ScatteringSimulation* Instrument2DItem::createScatteringSimulation(const MultiLayer& sample) const
{
    auto beam = beamItem()->createBeam();
    beam->setPolarization(m_polarization);
    auto detector = detectorItem()->createDetector();
    detector->setAnalyzer(m_analyzerDirection, m_analyzerEfficiency, m_analyzerTotalTransmission);
    detector->setDetectorNormal(beam->ki());

    return new ScatteringSimulation(*beam, sample, *detector);
}

OffspecSimulation* Instrument2DItem::createOffspecSimulation(const MultiLayer& sample) const
{
    auto beam = beamItem()->createBeam();
    beam->setPolarization(m_polarization);
    auto detector = detectorItem()->createDetector();
    detector->setAnalyzer(m_analyzerDirection, m_analyzerEfficiency, m_analyzerTotalTransmission);
    detector->setDetectorNormal(beam->ki());

    return new OffspecSimulation(*beam, sample, *detector);
}

//  ************************************************************************************************
//  class GISASInstrumentItem
//  ************************************************************************************************

std::vector<int> GISASInstrumentItem::shape() const
{
    auto* detector_item = detectorItem();
    return {detector_item->xSize(), detector_item->ySize()};
}

void GISASInstrumentItem::updateToRealData(const RealItem* item)
{
    if (!item)
        return;

    const auto data_shape = item->shape();
    if (shape().size() != data_shape.size())
        throw Error("Error in GISASInstrumentItem::updateToRealData: The type of "
                    "instrument is incompatible with passed data shape.");
    detectorItem()->setXSize(data_shape[0]);
    detectorItem()->setYSize(data_shape[1]);
}

QString GISASInstrumentItem::defaultName() const
{
    return "GISAS";
}

QString GISASInstrumentItem::instrumentType() const
{
    return "GISAS";
}

const ICoordSystem* GISASInstrumentItem::createCoordSystem() const
{
    const auto instrument = createInstrument();
    return instrument->detector().scatteringCoords(*instrument->beam());
}

//  ************************************************************************************************
//  class OffspecInstrumentItem
//  ************************************************************************************************

OffspecInstrumentItem::OffspecInstrumentItem()
{
    m_alphaAxis.initMin("Min", "Starting value", 0.0, Unit::degree, RealLimits::limited(-90, 90));
    m_alphaAxis.initMax("Max", "Ending value", 10.0, Unit::degree, RealLimits::limited(-90, 90));

    auto* beam_item = dynamic_cast<GISASBeamItem*>(beamItem());
    beam_item->setInclinationAngleGetter([=]() { return m_alphaAxis.min(); });
}

void OffspecInstrumentItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    Instrument2DItem::writeTo(w);
    w->writeEndElement();

    // alpha axis
    w->writeStartElement(Tag::AlphaAxis);
    m_alphaAxis.writeTo(w);
    w->writeEndElement();
}

void OffspecInstrumentItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            Instrument2DItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // alpha axis
        } else if (tag == Tag::AlphaAxis) {
            m_alphaAxis.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

std::vector<int> OffspecInstrumentItem::shape() const
{
    return {(int)m_alphaAxis.nbins(), detectorItem()->ySize()};
}

void OffspecInstrumentItem::updateToRealData(const RealItem* dataItem)
{
    if (!dataItem)
        return;

    const auto data_shape = dataItem->shape();
    if (shape().size() != data_shape.size())
        throw Error("Error in OffspecInstrumentItem::updateToRealData: The type of "
                    "instrument is incompatible with passed data shape.");

    m_alphaAxis.setNbins(data_shape[0]);
    detectorItem()->setYSize(data_shape[1]);
}

QString OffspecInstrumentItem::defaultName() const
{
    return "Offspec";
}

QString OffspecInstrumentItem::instrumentType() const
{
    return "Off specular";
}

const ICoordSystem* OffspecInstrumentItem::createCoordSystem() const
{
    IAxis* alphaAxis =
        new FixedBinAxis("alpha", m_alphaAxis.nbins(), m_alphaAxis.min() * Units::deg,
                         m_alphaAxis.max() * Units::deg);
    const auto instrument = createInstrument();
    return instrument->detector().offspecCoords(alphaAxis, instrument->beam()->ki());
}
