//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/MaskItems.cpp
//! @brief     Implements MaskItems classes
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/MaskItems.h"
#include "Device/Mask/Ellipse.h"
#include "Device/Mask/InfinitePlane.h"
#include "Device/Mask/Line.h"
#include "Device/Mask/Polygon.h"
#include "Device/Mask/Rectangle.h"
#include "GUI/Model/Model/SessionModel.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {
namespace Tag {

const QString Name("Name");
const QString MaskValue("MaskValue");
const QString IsVisible("IsVisible");
const QString BaseData("BaseData");
const QString XLow("XLow");
const QString YLow("YLow");
const QString XUp("XUp");
const QString YUp("YUp");
const QString IsClosed("IsClosed");
const QString PolygonPoint("PolygonPoint");
const QString XPosition("XPosition");
const QString YPosition("YPosition");
const QString XCenter("XCenter");
const QString YCenter("YCenter");
const QString XRadius("XRadius");
const QString YRadius("YRadius");
const QString Angle("Angle");
const QString Mask("Mask");

} // namespace Tag
} // namespace

MaskContainerItem::MaskContainerItem()
    : MaskItemObject(M_TYPE)
{
    const QString T_MASKS = "Mask Tag";
    static const QStringList allowedMasks = {RectangleItem::M_TYPE,       PolygonItem::M_TYPE,
                                             EllipseItem::M_TYPE,         VerticalLineItem::M_TYPE,
                                             HorizontalLineItem::M_TYPE,  MaskAllItem::M_TYPE,
                                             RegionOfInterestItem::M_TYPE};
    registerTag(T_MASKS, 0, -1, allowedMasks);
    setDefaultTag(T_MASKS);
}

QVector<MaskItem*> MaskContainerItem::maskItems() const
{
    QVector<MaskItem*> result;
    for (auto* child : itemChildren())
        if (auto* mask = dynamic_cast<MaskItem*>(child))
            result.append(mask);

    return result;
}

std::vector<SelectionProperty<MaskItemCatalog>> MaskContainerItem::maskItemSelections()
{
    std::vector<SelectionProperty<MaskItemCatalog>> maskSelections;
    for (auto* maskItem : maskItems()) {
        SelectionProperty<MaskItemCatalog> newMaskSelection;
        newMaskSelection.setCurrentItem(maskItem);
        maskSelections.push_back(std::move(newMaskSelection));
    }
    return maskSelections;
}

void MaskContainerItem::insertMask(int row, MaskItem* maskItem)
{
    insertChild(row, maskItem);
}

void MaskContainerItem::addMask(MaskItem* maskItem)
{
    insertChild(-1, maskItem);
}

RegionOfInterestItem* MaskContainerItem::regionOfInterestItem() const
{
    return firstChildOfType<RegionOfInterestItem>();
}

void MaskContainerItem::clear()
{
    model()->removeRows(0, numberOfChildren(), index());
}

QString MaskItem::maskName() const
{
    return itemName();
}

void MaskItem::setMaskName(const QString& name)
{
    setItemName(name);
}

/* ------------------------------------------------------------------------- */

bool MaskItem::maskValue() const
{
    return getItemValue(P_MASK_VALUE).toBool();
}

void MaskItem::setMaskValue(bool mask_value)
{
    setItemValue(P_MASK_VALUE, mask_value);
    emit maskVisibilityChanged();
}

bool MaskItem::isVisibleValue() const
{
    return getItemValue(P_IS_VISIBLE).toBool();
}

void MaskItem::setIsVisibleValue(bool visible)
{
    setItemValue(P_IS_VISIBLE, visible);
    emit maskVisibilityChanged();
}

void MaskItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // name
    w->writeStartElement(Tag::Name);
    XML::writeAttribute(w, XML::Attrib::value, maskName());
    w->writeEndElement();

    // mask value
    w->writeStartElement(Tag::MaskValue);
    XML::writeAttribute(w, XML::Attrib::value, maskValue());
    w->writeEndElement();

    // is visible?
    w->writeStartElement(Tag::IsVisible);
    XML::writeAttribute(w, XML::Attrib::value, isVisibleValue());
    w->writeEndElement();
}

void MaskItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // name
        if (tag == Tag::Name) {
            QString name;
            XML::readAttribute(r, XML::Attrib::value, &name);
            setMaskName(name);
            XML::gotoEndElementOfTag(r, tag);

            // mask value
        } else if (tag == Tag::MaskValue) {
            bool maskValue = true;
            XML::readAttribute(r, XML::Attrib::value, &maskValue);
            setMaskValue(maskValue);
            XML::gotoEndElementOfTag(r, tag);

            // is visible?
        } else if (tag == Tag::IsVisible) {
            bool isVisible = true;
            XML::readAttribute(r, XML::Attrib::value, &isVisible);
            setIsVisibleValue(isVisible);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

MaskItem::MaskItem(const QString& name)
    : MaskItemObject(name)
{
    addProperty(P_MASK_VALUE, true);
    addProperty(P_IS_VISIBLE, true);
}

SessionItem* MaskItem::maskValueItem() const
{
    return getItem(P_MASK_VALUE);
}

/* ------------------------------------------------------------------------- */

RectangleItem::RectangleItem()
    : RectangleItem(M_TYPE)
{
}

RectangleItem::RectangleItem(const QString& modelType)
    : MaskItem(modelType)
{
    setItemName(modelType);
    addProperty(P_XLOW, double(0.0));
    addProperty(P_YLOW, double(0.0));
    addProperty(P_XUP, double(0.0));
    addProperty(P_YUP, double(0.0));
}

std::unique_ptr<IShape2D> RectangleItem::createShape(double scale) const
{
    double xlow = scale * xLow();
    double ylow = scale * yLow();
    double xup = scale * xUp();
    double yup = scale * yUp();
    return std::make_unique<Rectangle>(xlow, ylow, xup, yup);
}

double RectangleItem::xLow() const
{
    return getItemValue(P_XLOW).toDouble();
}

void RectangleItem::setXLow(double val)
{
    setItemValue(P_XLOW, val);
    emit maskGeometryChanged();
}

double RectangleItem::yLow() const
{
    return getItemValue(P_YLOW).toDouble();
}

void RectangleItem::setYLow(double val)
{
    setItemValue(P_YLOW, val);
    emit maskGeometryChanged();
}

double RectangleItem::xUp() const
{
    return getItemValue(P_XUP).toDouble();
}

void RectangleItem::setXUp(double val)
{
    setItemValue(P_XUP, val);
    emit maskGeometryChanged();
}

double RectangleItem::yUp() const
{
    return getItemValue(P_YUP).toDouble();
}

void RectangleItem::setYUp(double val)
{
    setItemValue(P_YUP, val);
    emit maskGeometryChanged();
}

void RectangleItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    MaskItem::writeTo(w);
    w->writeEndElement();

    // x low
    w->writeStartElement(Tag::XLow);
    XML::writeAttribute(w, XML::Attrib::value, xLow());
    w->writeEndElement();

    // y low
    w->writeStartElement(Tag::YLow);
    XML::writeAttribute(w, XML::Attrib::value, yLow());
    w->writeEndElement();

    // x up
    w->writeStartElement(Tag::XUp);
    XML::writeAttribute(w, XML::Attrib::value, xUp());
    w->writeEndElement();

    // y up
    w->writeStartElement(Tag::YUp);
    XML::writeAttribute(w, XML::Attrib::value, yUp());
    w->writeEndElement();
}

void RectangleItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            MaskItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // x low
        } else if (tag == Tag::XLow) {
            double xLow = 0;
            XML::readAttribute(r, XML::Attrib::value, &xLow);
            setXLow(xLow);
            XML::gotoEndElementOfTag(r, tag);

            // y low
        } else if (tag == Tag::YLow) {
            double yLow = 0;
            XML::readAttribute(r, XML::Attrib::value, &yLow);
            setYLow(yLow);
            XML::gotoEndElementOfTag(r, tag);

            // x up
        } else if (tag == Tag::XUp) {
            double xUp = 0;
            XML::readAttribute(r, XML::Attrib::value, &xUp);
            setXUp(xUp);
            XML::gotoEndElementOfTag(r, tag);

            // y up
        } else if (tag == Tag::YUp) {
            double yUp = 0;
            XML::readAttribute(r, XML::Attrib::value, &yUp);
            setYUp(yUp);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

/* ------------------------------------------------------------------------- */

RegionOfInterestItem::RegionOfInterestItem()
    : RectangleItem(M_TYPE)
{
    setMaskValue(false);
}

std::unique_ptr<IShape2D> RegionOfInterestItem::createShape(double scale) const
{
    auto shape = RectangleItem::createShape(scale);
    dynamic_cast<Rectangle*>(shape.get())->setInverted(true);
    return shape;
}

/* ------------------------------------------------------------------------- */

PolygonPointItem::PolygonPointItem()
    : MaskItemObject(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_POSX, double(0.0));
    addProperty(P_POSY, double(0.0));
}

double PolygonPointItem::posX() const
{
    return getItemValue(P_POSX).toDouble();
}

void PolygonPointItem::setPosX(double val)
{
    setItemValue(P_POSX, val);
    emit maskGeometryChanged();
}

double PolygonPointItem::posY() const
{
    return getItemValue(P_POSY).toDouble();
}

void PolygonPointItem::setPosY(double val)
{
    setItemValue(P_POSY, val);
    emit maskGeometryChanged();
}

void PolygonPointItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::xPos, posX());
    XML::writeAttribute(w, XML::Attrib::yPos, posY());
    XML::writeAttribute(w, XML::Attrib::version, uint(1));
}

void PolygonPointItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    double xPos = 0;
    XML::readAttribute(r, XML::Attrib::xPos, &xPos);
    setPosX(xPos);

    double yPos = 0;
    XML::readAttribute(r, XML::Attrib::yPos, &yPos);
    setPosY(yPos);
}

/* ------------------------------------------------------------------------- */

PolygonItem::PolygonItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    const QString T_POINTS = "Point tag";
    registerTag(T_POINTS, 0, -1, {PolygonPointItem::M_TYPE});
    setDefaultTag(T_POINTS);
    addProperty(P_ISCLOSED, false);
}

std::unique_ptr<IShape2D> PolygonItem::createShape(double scale) const
{
    std::vector<double> x, y;
    for (auto* item : this->childrenOfType<PolygonPointItem>()) {
        x.push_back(scale * item->posX());
        y.push_back(scale * item->posY());
    }
    return std::make_unique<Polygon>(x, y);
}

bool PolygonItem::isClosed() const
{
    return getItemValue(P_ISCLOSED).toBool();
}

void PolygonItem::setIsClosed(bool closed)
{
    setItemValue(P_ISCLOSED, closed);
}

QVector<PolygonPointItem*> PolygonItem::points() const
{
    return childrenOfType<PolygonPointItem>();
}

void PolygonItem::addPoint(double x, double y)
{
    auto* pointItem = new PolygonPointItem();
    pointItem->setPosX(x);
    pointItem->setPosY(y);
    insertChild(-1, pointItem);
}

void PolygonItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    MaskItem::writeTo(w);
    w->writeEndElement();

    // is closed?
    w->writeStartElement(Tag::IsClosed);
    XML::writeAttribute(w, XML::Attrib::value, isClosed()); // TODO do we need to save 'isClosed'?
    w->writeEndElement();

    // polygon points
    for (const auto* p : points()) {
        w->writeStartElement(Tag::PolygonPoint);
        p->writeTo(w);
        w->writeEndElement();
    }
}

void PolygonItem::readFrom(QXmlStreamReader* r)
{
    // m_points.clear(); do not forget to empty owning vector before filling it with new objects

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            MaskItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // is closed?
        } else if (tag == Tag::IsClosed) {
            bool isClosed = false;
            XML::readAttribute(r, XML::Attrib::value,
                               &isClosed); // TODO do we need to read 'isClosed'?
            setIsClosed(isClosed);
            XML::gotoEndElementOfTag(r, tag);

            // polygon point
        } else if (tag == Tag::PolygonPoint) {
            PolygonPointItem point;
            point.readFrom(r);
            addPoint(point.posX(), point.posY());
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

/* ------------------------------------------------------------------------- */

VerticalLineItem::VerticalLineItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_POSX, double(0.0));
}

std::unique_ptr<IShape2D> VerticalLineItem::createShape(double scale) const
{
    return std::make_unique<VerticalLine>(scale * posX());
}

double VerticalLineItem::posX() const
{
    return getItemValue(P_POSX).toDouble();
}

void VerticalLineItem::setPosX(double val)
{
    setItemValue(P_POSX, val);
    emit maskGeometryChanged(this);
}

void VerticalLineItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    MaskItem::writeTo(w);
    w->writeEndElement();

    // x position
    w->writeStartElement(Tag::XPosition);
    XML::writeAttribute(w, XML::Attrib::value, posX());
    w->writeEndElement();
}

void VerticalLineItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            MaskItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // x position
        } else if (tag == Tag::XPosition) {
            double xPos = 0;
            XML::readAttribute(r, XML::Attrib::value, &xPos);
            setPosX(xPos);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

/* ------------------------------------------------------------------------- */

HorizontalLineItem::HorizontalLineItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_POSY, double(0.0));
}

std::unique_ptr<IShape2D> HorizontalLineItem::createShape(double scale) const
{
    return std::make_unique<HorizontalLine>(scale * posY());
}

double HorizontalLineItem::posY() const
{
    return getItemValue(P_POSY).toDouble();
}

void HorizontalLineItem::setPosY(double pos_y)
{
    setItemValue(P_POSY, pos_y);
    emit maskGeometryChanged(this);
}

void HorizontalLineItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    MaskItem::writeTo(w);
    w->writeEndElement();

    // y position
    w->writeStartElement(Tag::YPosition);
    XML::writeAttribute(w, XML::Attrib::value, posY());
    w->writeEndElement();
}

void HorizontalLineItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            MaskItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // y position
        } else if (tag == Tag::YPosition) {
            double yPos = 0;
            XML::readAttribute(r, XML::Attrib::value, &yPos);
            setPosY(yPos);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

/* ------------------------------------------------------------------------- */

EllipseItem::EllipseItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    addProperty(P_XCENTER, double(0.0));
    addProperty(P_YCENTER, double(0.0));
    addProperty(P_XRADIUS, double(0.0));
    addProperty(P_YRADIUS, double(0.0));
    addProperty(P_ANGLE, double(0.0));
}

std::unique_ptr<IShape2D> EllipseItem::createShape(double scale) const
{
    double xcenter = scale * xCenter();
    double ycenter = scale * yCenter();
    double xradius = scale * xRadius();
    double yradius = scale * yRadius();
    double _angle = scale * angle();

    return std::make_unique<Ellipse>(xcenter, ycenter, xradius, yradius, _angle);
}

double EllipseItem::xCenter() const
{
    return getItemValue(P_XCENTER).toDouble();
}

void EllipseItem::setXCenter(double val)
{
    setItemValue(P_XCENTER, val);
    emit maskGeometryChanged();
}

double EllipseItem::yCenter() const
{
    return getItemValue(P_YCENTER).toDouble();
}

void EllipseItem::setYCenter(double val)
{
    setItemValue(P_YCENTER, val);
    emit maskGeometryChanged();
}

double EllipseItem::xRadius() const
{
    return getItemValue(P_XRADIUS).toDouble();
}

void EllipseItem::setXRadius(double val)
{
    setItemValue(P_XRADIUS, val);
    emit maskGeometryChanged();
}

double EllipseItem::yRadius() const
{
    return getItemValue(P_YRADIUS).toDouble();
}

void EllipseItem::setYRadius(const double val)
{
    setItemValue(P_YRADIUS, val);
    emit maskGeometryChanged();
}

double EllipseItem::angle() const
{
    return getItemValue(P_ANGLE).toDouble();
}

void EllipseItem::setAngle(const double val)
{
    setItemValue(P_ANGLE, val);
    emit maskGeometryChanged();
}

void EllipseItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    MaskItem::writeTo(w);
    w->writeEndElement();

    // x center
    w->writeStartElement(Tag::XCenter);
    XML::writeAttribute(w, XML::Attrib::value, xCenter());
    w->writeEndElement();

    // y center
    w->writeStartElement(Tag::YCenter);
    XML::writeAttribute(w, XML::Attrib::value, yCenter());
    w->writeEndElement();

    // x radius
    w->writeStartElement(Tag::XRadius);
    XML::writeAttribute(w, XML::Attrib::value, xRadius());
    w->writeEndElement();

    // y radius
    w->writeStartElement(Tag::YRadius);
    XML::writeAttribute(w, XML::Attrib::value, yRadius());
    w->writeEndElement();

    // rotation angle
    w->writeStartElement(Tag::Angle);
    XML::writeAttribute(w, XML::Attrib::value, angle());
    w->writeEndElement();
}

void EllipseItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            MaskItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // x center
        } else if (tag == Tag::XCenter) {
            double xCenter = 0;
            XML::readAttribute(r, XML::Attrib::value, &xCenter);
            setXCenter(xCenter);
            XML::gotoEndElementOfTag(r, tag);

            // y center
        } else if (tag == Tag::YCenter) {
            double yCenter = 0;
            XML::readAttribute(r, XML::Attrib::value, &yCenter);
            setYCenter(yCenter);
            XML::gotoEndElementOfTag(r, tag);

            // x radius
        } else if (tag == Tag::XRadius) {
            double xRadius = 0;
            XML::readAttribute(r, XML::Attrib::value, &xRadius);
            setXRadius(xRadius);
            XML::gotoEndElementOfTag(r, tag);

            // y radius
        } else if (tag == Tag::YRadius) {
            double yRadius = 0;
            XML::readAttribute(r, XML::Attrib::value, &yRadius);
            setYRadius(yRadius);
            XML::gotoEndElementOfTag(r, tag);

            // rotation angle
        } else if (tag == Tag::Angle) {
            double angle = 0;
            XML::readAttribute(r, XML::Attrib::value, &angle);
            setAngle(angle);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

/* ------------------------------------------------------------------------- */

MaskAllItem::MaskAllItem()
    : MaskItem(M_TYPE)
{
    setItemName(M_TYPE);
    maskValueItem()->setEnabled(false);
}

std::unique_ptr<IShape2D> MaskAllItem::createShape(double) const
{
    return std::make_unique<InfinitePlane>();
}

/* ------------------------------------------------------------------------- */

MaskItems::MaskItems()
    : SessionModel("MaskItems")
{
    m_maskContainer = insertItem<MaskContainerItem>();
}

void MaskItems::insertMask(int index, MaskItem* maskItem)
{
    m_maskContainer->insertChild(index, maskItem);
}

QVector<MaskItem*> MaskItems::maskItems() const
{
    return m_maskContainer->maskItems();
}

void MaskItems::copy(const MaskContainerItem* maskContainer)
{
    m_maskContainer->clear();
    if (maskContainer)
        for (auto mask : maskContainer->maskItems())
            copyItem(mask, m_maskContainer);
}

void MaskItems::clear()
{
    SessionModel::clear();
    m_maskContainer = insertItem<MaskContainerItem>();
}

void MaskItems::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // masks
    for (const auto& sel : m_maskContainer->maskItemSelections()) {
        w->writeStartElement(Tag::Mask);
        sel.writeTo(w);
        w->writeEndElement();
    }
}

void MaskItems::readFrom(QXmlStreamReader* r, MessageService*)
{
    m_maskContainer->clear();

    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // mask
        if (tag == Tag::Mask) {
            SelectionProperty<MaskItemCatalog> sel;
            sel.readFrom(r);
            m_maskContainer->addMask(sel.currentItem());
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

MaskItemObject::MaskItemObject(const QString& modelType)
    : SessionItem(modelType)
{
}

MaskItemObject::~MaskItemObject()
{
    emit maskToBeDestroyed(this);
}
