//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Device/SpecularBeamInclinationItem.cpp
//! @brief     Implements class SpecularBeamInclinationItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Device/SpecularBeamInclinationItem.h"
#include "Base/Axis/IAxis.h"
#include "Base/Const/Units.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/CatDevice/DistributionItemCatalog.h"
#include "GUI/Model/Device/PointwiseAxisItem.h"
#include "GUI/Support/XML/UtilXML.h"

namespace {
namespace Tag {

const QString UniformAxis("UniformAxis");
const QString PointwiseAxis("PointwiseAxis");
const QString BeamInclinationDistribution("BeamInclinationDistribution");
const QString IsUniformAxis("IsUniformAxis");
const QString BaseData("BaseData");

} // namespace Tag

void setAxisPresentationDefaults(BasicAxisItem* axisItem)
{
    ASSERT(axisItem);
    axisItem->setTitle("alpha_i");

    if (!dynamic_cast<PointwiseAxisItem*>(axisItem)) {
        axisItem->setMin(0.0);
        axisItem->setMax(3.0);
        axisItem->setBinCount(500);
    }
}

void initDistribution(DistributionItem* newDistribution,
                      const DistributionItem* currentDistribution)
{
    newDistribution->setLimits(RealLimits::limited(-90.0, 90.0));
    newDistribution->setUnit(Unit::degree);

    double meanVal = 0.0;
    if (auto* cd = dynamic_cast<const SymmetricResolutionItem*>(currentDistribution))
        meanVal = cd->mean();

    if (auto* nd = dynamic_cast<SymmetricResolutionItem*>(newDistribution)) {
        nd->setMean(meanVal);
        nd->setMeanDecimals(3);
    }

    newDistribution->initDistribution(meanVal);
}

} // namespace

SpecularBeamInclinationItem::SpecularBeamInclinationItem(const InstrumentItem* owningInstrument)
    : BeamDistributionItem()
    , m_owningInstrument(owningInstrument)
{
    m_distribution.initWithInitializer("Distribution", "",
                                       DistributionItemCatalog::symmetricTypes(), initDistribution);

    m_uniformAlphaAxis.reset(new BasicAxisItem());
    setAxisPresentationDefaults(m_uniformAlphaAxis.get());
    m_currentAxisIsUniformAxis = true;
}

void SpecularBeamInclinationItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    BeamDistributionItem::writeTo(w);
    w->writeEndElement();

    // is uniform axis?
    w->writeStartElement(Tag::IsUniformAxis);
    XML::writeAttribute(w, XML::Attrib::value, m_currentAxisIsUniformAxis);
    w->writeEndElement();

    // uniform axis
    if (m_uniformAlphaAxis) {
        w->writeStartElement(Tag::UniformAxis);
        m_uniformAlphaAxis->writeTo(w);
        w->writeEndElement();
    }

    // pointwise axis
    if (m_pointwiseAlphaAxis) {
        w->writeStartElement(Tag::PointwiseAxis);
        m_pointwiseAlphaAxis->writeTo(w);
        w->writeEndElement();
    }
}

void SpecularBeamInclinationItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            BeamDistributionItem::readFrom(r);
            m_distribution->setLimits(RealLimits::limited(-90.0, 90.0));
            m_distribution->setUnit(Unit::degree);
            XML::gotoEndElementOfTag(r, tag);

            // is uniform axis?
        } else if (tag == Tag::IsUniformAxis) {
            XML::readAttribute(r, XML::Attrib::value, &m_currentAxisIsUniformAxis);
            XML::gotoEndElementOfTag(r, tag);

            // uniform axis
        } else if (tag == Tag::UniformAxis) {
            m_uniformAlphaAxis = std::make_unique<BasicAxisItem>();
            setAxisPresentationDefaults(m_uniformAlphaAxis.get());
            m_uniformAlphaAxis->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // pointwise axis
        } else if (tag == Tag::PointwiseAxis) {
            m_pointwiseAlphaAxis = std::make_unique<PointwiseAxisItem>();
            setAxisPresentationDefaults(m_pointwiseAlphaAxis.get());
            m_pointwiseAlphaAxis->readFrom(r);
            m_pointwiseAlphaAxis->setOwningInstrument(m_owningInstrument);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

double SpecularBeamInclinationItem::scaleFactor() const
{
    return Units::deg;
}

BasicAxisItem* SpecularBeamInclinationItem::alphaAxisItem() const
{
    return m_currentAxisIsUniformAxis ? m_uniformAlphaAxis.get() : m_pointwiseAlphaAxis.get();
}

bool SpecularBeamInclinationItem::pointwiseAlphaAxisDefined() const
{
    return m_pointwiseAlphaAxis.get() != nullptr;
}

bool SpecularBeamInclinationItem::pointwiseAlphaAxisSelected() const
{
    return !m_currentAxisIsUniformAxis;
}

bool SpecularBeamInclinationItem::uniformAlphaAxisSelected() const
{
    return m_currentAxisIsUniformAxis;
}

void SpecularBeamInclinationItem::selectUniformAxis()
{
    m_currentAxisIsUniformAxis = true;
}

void SpecularBeamInclinationItem::selectPointwiseAxis()
{
    ASSERT(pointwiseAlphaAxisDefined());
    m_currentAxisIsUniformAxis = false;
}

void SpecularBeamInclinationItem::initUniformAxis(const IAxis& axis)
{
    m_uniformAlphaAxis->setBinCount(static_cast<int>(axis.size()));
}

void SpecularBeamInclinationItem::initPointwiseAxis(const IAxis& axis, QString units)
{
    if (!m_pointwiseAlphaAxis) {
        m_pointwiseAlphaAxis.reset(new PointwiseAxisItem());
        m_pointwiseAlphaAxis->setOwningInstrument(m_owningInstrument);
        setAxisPresentationDefaults(m_pointwiseAlphaAxis.get());
    }

    m_pointwiseAlphaAxis->init(axis, units);
    m_pointwiseAlphaAxis->updateIndicators();
}
