//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Project/ProjectDocument.cpp
//! @brief     Implements class ProjectDocument
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Data/RealItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Model/JobModel.h"
#include "GUI/Model/Model/RealModel.h"
#include "GUI/Model/Project/IOService.h"
#include "GUI/Model/Project/LinkInstrumentManager.h"
#include "GUI/Model/Sample/ItemWithMaterial.h"
#include "GUI/Model/Sample/MaterialItem.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Support/Data/SimulationOptionsItem.h"
#include "GUI/Support/IO/ProjectUtils.h"
#include "GUI/Support/XML/UtilXML.h"
#include "GUI/Util/DeserializationException.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/MessageService.h"
#include "GUI/Util/Path.h"
#include <QStandardPaths>
#include <QXmlStreamReader>
#include <memory>

std::optional<ProjectDocument*> gProjectDocument;

namespace {

const QString minimal_supported_version = "1.6.0";

namespace Tag {

const QString BornAgain("BornAgain");
const QString DocumentInfo("DocumentInfo");
const QString SimulationOptions("SimulationOptions");
const QString SampleModel("SampleModel");
const QString InstrumentModel("InstrumentModel");

} // namespace Tag
} // namespace


ProjectDocument::ProjectDocument()
    : m_modified(false)
    , m_dataService(new DatafieldIOService(this))
    , m_singleInstrumentMode(false)
    , m_singleSampleMode(false)
    , m_functionalities(All)
    , m_instrumentEditController(&m_instrumentModel)
{
    connect(&m_instrumentEditController, &InstrumentsEditController::instrumentAddedOrRemoved, this,
            &ProjectDocument::onModelChanged, Qt::UniqueConnection);
    connect(&m_instrumentEditController, &InstrumentsEditController::instrumentChanged, this,
            &ProjectDocument::onModelChanged, Qt::UniqueConnection);

    m_linkManager = std::make_unique<LinkInstrumentManager>(this);
    setObjectName("ProjectDocument");

    m_dataService->setApplicationModels(&m_applicationModels);
    connectModels();
}

QString ProjectDocument::projectName() const
{
    return m_project_name;
}

void ProjectDocument::setProjectName(const QString& text)
{
    if (m_project_name != text)
        m_project_name = text;
}

QString ProjectDocument::projectDir() const
{
    return m_project_dir;
}

QString ProjectDocument::validProjectDir() const
{
    if (m_project_name.isEmpty())
        return "";
    return m_project_dir;
}

void ProjectDocument::setProjectDir(const QString& text)
{
    m_project_dir = text;
}

//! Returns directory name suitable for saving plots.

QString ProjectDocument::userExportDir() const
{
    if (QString dir = validProjectDir(); !dir.isEmpty())
        return dir;
    return QStandardPaths::writableLocation(QStandardPaths::HomeLocation);
}

QString ProjectDocument::projectFileName() const
{
    if (!projectName().isEmpty())
        return projectDir() + "/" + projectName() + GUI::Project::Utils::projectFileExtension;
    return "";
}

void ProjectDocument::setProjectFileName(const QString& projectFileName)
{
    setProjectName(GUI::Project::Utils::projectName(projectFileName));
    setProjectDir(GUI::Project::Utils::projectDir(projectFileName));
}

InstrumentModel* ProjectDocument::instrumentModel() const
{
    return const_cast<InstrumentModel*>(&m_instrumentModel);
}

SampleModel* ProjectDocument::sampleModel()
{
    return &m_sampleModel;
}

RealModel* ProjectDocument::realModel() const
{
    return m_applicationModels.realModel();
}

JobModel* ProjectDocument::jobModel() const
{
    return m_applicationModels.jobModel();
}

SimulationOptionsItem* ProjectDocument::simulationOptionsItem()
{
    return &m_simulationOptionsItem;
}

LinkInstrumentManager* ProjectDocument::linkInstrumentManager()
{
    return m_linkManager.get();
}

InstrumentsEditController* ProjectDocument::instrumentsEditController()
{
    return &m_instrumentEditController;
}

// This function is only used in tests
void ProjectDocument::save(const QString& project_file_name, bool autoSave)
{
    saveProjectData(project_file_name);
    saveProjectFile(project_file_name, autoSave);
}

void ProjectDocument::saveProjectFile(const QString& project_file_name, bool autoSave)
{
    QFile file(project_file_name);
    if (!file.open(QFile::ReadWrite | QIODevice::Truncate | QFile::Text))
        throw Error("ProjectDocument::save_project_file() -> Error. Cannot open "
                    "file '"
                    + project_file_name + "' for writing.");

    writeProject(&file);
    file.close();

    if (!autoSave) {
        setProjectFileName(project_file_name);
        clearModified();
    }
}

void ProjectDocument::saveProjectData(const QString& project_file_name)
{
    m_dataService->saveDataFiles(GUI::Project::Utils::projectDir(project_file_name));
}

ProjectDocument::ReadResult ProjectDocument::loadProjectFile(const QString& project_file_name,
                                                             MessageService& messageService)
{
    setProjectFileName(project_file_name);

    QFile file(projectFileName());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QString message = QString("Open file error '%1'").arg(file.errorString());
        messageService.addError(this, message);
        return ReadResult::error;
    }

    try {
        disconnectModels();
        auto result = readProject(&file, messageService);
        file.close();
        if (result == ReadResult::error)
            return result;

        m_dataService->loadDataFiles(projectDir(), &messageService);
        connectModels();

        if (!messageService.warnings().empty())
            result = ReadResult::warning;
        return result;
    } catch (const std::exception& ex) {
        QString message = QString("Exception thrown '%1'").arg(QString(ex.what()));
        messageService.addError(this, message);
        return ReadResult::error;
    }
}

bool ProjectDocument::hasValidNameAndPath()
{
    return (!m_project_name.isEmpty() && !m_project_dir.isEmpty());
}

bool ProjectDocument::isModified() const
{
    return m_modified;
}

void ProjectDocument::setModified()
{
    m_modified = true;
    emit modifiedStateChanged();
}

void ProjectDocument::clearModified()
{
    m_modified = false;
    emit modifiedStateChanged();
}

bool ProjectDocument::hasData() const
{
    return !m_dataService->dataInterfaces().isEmpty();
}

QString ProjectDocument::documentVersion() const
{
    QString result(m_currentVersion);
    if (result.isEmpty())
        result = GUI::Util::Path::getBornAgainVersionString();
    return result;
}

bool ProjectDocument::singleInstrumentMode() const
{
    return m_singleInstrumentMode;
}

void ProjectDocument::setSingleInstrumentMode(bool b)
{
    if (b != m_singleInstrumentMode) {
        m_singleInstrumentMode = b;
        emit singleInstrumentModeChanged();
        setModified();
    }
}

bool ProjectDocument::singleSampleMode() const
{
    return m_singleSampleMode;
}

void ProjectDocument::setSingleSampleMode(bool b)
{
    if (b != m_singleSampleMode) {
        m_singleSampleMode = b;
        emit singleSampleModeChanged();
        setModified();
    }
}

ProjectDocument::Functionalities ProjectDocument::functionalities() const
{
    return m_functionalities;
}

void ProjectDocument::setFunctionalities(const Functionalities& f)
{
    if (m_functionalities != f) {
        m_functionalities = f;
        emit functionalitiesChanged();
        setModified();
    }
}

void ProjectDocument::onModelChanged()
{
    setModified();
}

void ProjectDocument::writeProject(QIODevice* device)
{
    QXmlStreamWriter w(device);
    w.setAutoFormatting(true);
    w.writeStartDocument();
    w.writeStartElement(Tag::BornAgain);
    QString version_string = GUI::Util::Path::getBornAgainVersionString();
    w.writeAttribute(XML::Attrib::BA_Version, version_string);
    XML::writeAttribute(&w, XML::Attrib::version, uint(2));

    w.writeStartElement(Tag::DocumentInfo);
    w.writeAttribute(XML::Attrib::projectName, projectName());
    w.writeEndElement();

    // simulation options
    w.writeStartElement(Tag::SimulationOptions);
    m_simulationOptionsItem.writeTo(&w);
    w.writeEndElement();

    // instruments
    w.writeStartElement(Tag::InstrumentModel);
    m_instrumentModel.writeTo(&w);
    w.writeEndElement();

    // samples
    w.writeStartElement(Tag::SampleModel);
    m_sampleModel.writeTo(&w);
    w.writeEndElement();

    // job model and real model
    m_applicationModels.writeTo(&w);

    w.writeEndElement(); // BornAgain tag
    w.writeEndDocument();
}

ProjectDocument::ReadResult ProjectDocument::readProject(QIODevice* device,
                                                         MessageService& messageService)
{
    const int warningsBefore = messageService.warnings().size();

    QXmlStreamReader r(device);
    try {
        while (!r.atEnd()) {
            r.readNext();
            if (r.isStartElement()) {
                if (r.name() == Tag::BornAgain) {
                    const uint version = XML::readUIntAttribute(&r, XML::Attrib::version);
                    Q_UNUSED(version)
                    m_currentVersion = r.attributes().value(XML::Attrib::BA_Version).toString();

                    if (!GUI::Util::Path::isVersionMatchMinimal(m_currentVersion,
                                                                minimal_supported_version)) {
                        QString message = QString("Cannot open document version '%1', "
                                                  "minimal supported version '%2'")
                                              .arg(m_currentVersion)
                                              .arg(minimal_supported_version);
                        messageService.addError(this, message);
                        return ReadResult::error;
                    }

                    while (r.readNextStartElement()) {
                        QString tag = r.name().toString();

                        if (tag == Tag::DocumentInfo) {
                            r.skipCurrentElement();

                            // simulation options
                        } else if (tag == Tag::SimulationOptions) {
                            m_simulationOptionsItem.readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // instruments
                        } else if (tag == Tag::InstrumentModel) {
                            m_instrumentModel.readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // samples
                        } else if (tag == Tag::SampleModel) {
                            m_sampleModel.readFrom(&r);
                            XML::gotoEndElementOfTag(&r, tag);

                            // job model and real model
                        } else
                            m_applicationModels.readFrom(&r, &messageService);
                    }
                }
            }
        }

    } catch (DeserializationException& ex) {
        r.raiseError(ex.text());
    }

    if (r.hasError()) {
        QString message = QString("Format error '%1'").arg(r.errorString());
        messageService.addError(this, message);
        return ReadResult::error;
    }

    // return "ReadResult::warning" only if warnings have been issued _in here_
    return warningsBefore != messageService.warnings().size() ? ReadResult::warning
                                                              : ReadResult::ok;
}

void ProjectDocument::disconnectModels()
{
    disconnect(&m_applicationModels, &ApplicationModels::modelChanged, this,
               &ProjectDocument::onModelChanged);
}

void ProjectDocument::connectModels()
{
    connect(&m_applicationModels, &ApplicationModels::modelChanged, this,
            &ProjectDocument::onModelChanged, Qt::UniqueConnection);
}
