//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Computation/DWBAComputation.cpp
//! @brief     Implements class DWBAComputation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Computation/DWBAComputation.h"
#include "Base/Progress/ProgressHandler.h"
#include "Resample/Element/DiffuseElement.h"
#include "Resample/Flux/IFlux.h"
#include "Resample/Options/SimulationOptions.h"
#include "Resample/Processed/ReLayout.h"
#include "Resample/Processed/ReSample.h"
#include "Sim/Contrib/DWBAContribution.h"
#include "Sim/Contrib/GISASSpecularContribution.h"
#include "Sim/Contrib/RoughMultiLayerContribution.h"

DWBAComputation::DWBAComputation(const ReSample& re_sample, const SimulationOptions& options,
                                 ProgressHandler& progress,
                                 std::vector<DiffuseElement>::iterator begin_it,
                                 std::vector<DiffuseElement>::iterator end_it)
    : IComputation(re_sample, options, progress)
    , m_begin_it(begin_it)
    , m_end_it(end_it)
{
}

DWBAComputation::~DWBAComputation() = default;

// The normalization of the calculated scattering intensities is:
// For nanoparticles: rho * (scattering cross-section/scattering particle)
// For roughness: (scattering cross-section of area S)/S
// For specular peak: |R|^2 * sin(alpha_i) / solid_angle
// This allows them to be added and normalized together to the beam afterwards
void DWBAComputation::runProtected()
{
    for (auto it = m_begin_it; it != m_end_it; ++it) {
        DiffuseElement& ele = *it;

        const Fluxes fluxes_in = m_re_sample.fluxesIn(ele.getKi());
        const Fluxes fluxes_out = m_re_sample.fluxesOut(ele.meanKf());
        ele.setFluxes(&fluxes_in, &fluxes_out);

        for (const ReLayout* relayout : m_re_sample.relayouts())
            compute::dwbaContribution(*relayout, ele);

        if (m_re_sample.hasRoughness())
            compute::roughMultiLayerContribution(m_re_sample, ele);

        if (m_options.includeSpecular())
            compute::gisasSpecularContribution(ele);

        stepProgress();
    }
}
