//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Scan/ISpecularScan.h
//! @brief     Declares interface ISpecularScan.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SCAN_ISPECULARSCAN_H
#define BORNAGAIN_SIM_SCAN_ISPECULARSCAN_H

#include "Base/Types/ICloneable.h"
#include <cstddef>
#include <heinz/Vectors3D.h>
#include <memory>
#include <vector>

class CoordSystem1D;
class IAxis;
class IFootprintFactor;
class PolFilter;
class SpecularElement;
class SpinMatrix;

//! Abstract base class for all types of specular scans.

class ISpecularScan : public ICloneable {
public:
    ISpecularScan* clone() const override = 0;
    ~ISpecularScan() override;

    //! Sets the polarization density matrix according to the given Bloch vector
    void setPolarization(R3 bloch_vector);

    //! Sets the polarization analyzer characteristics of the detector
    void setAnalyzer(R3 direction, double efficiency, double total_transmission);

    const PolFilter* analyzer() const { return m_polAnalyzer.get(); }

    virtual double wavelength() const = 0;

#ifndef SWIG
    SpinMatrix polarizerMatrix() const;
    SpinMatrix analyzerMatrix() const;

    //! Generates simulation elements for specular simulations
    virtual std::vector<SpecularElement> generateElements() const = 0;

    //! Returns coordinate axis assigned to the data holder
    virtual const IAxis* coordinateAxis() const = 0;

    //! Returns IFootprintFactor object pointer
    virtual const IFootprintFactor* footprintFactor() const = 0;

    //! Returns footprint correction factor for a range of simulation elements of size _n_elements_
    //! and starting from element with index _i_.
    virtual std::vector<double> footprint(size_t i, size_t n_elements) const = 0;

    //! Returns the number of simulation elements
    virtual size_t nSteps() const = 0;

    virtual CoordSystem1D* scanCoordSystem() const = 0;

    //! Returns intensity vector corresponding to convolution of given simulation elements
    virtual std::vector<double>
    createIntensities(const std::vector<SpecularElement>& eles) const = 0;

    bool polarized() const;

protected:
    std::unique_ptr<R3> m_beamPolarization; //!< Bloch vector encoding the beam's polarization
    std::unique_ptr<PolFilter> m_polAnalyzer;

#endif // SWIG
};

#endif // BORNAGAIN_SIM_SCAN_ISPECULARSCAN_H
