//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Scan/ScanResolution.cpp
//! @brief     Implements scan resolution class and derived ones.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Scan/ScanResolution.h"
#include "Param/Distrib/ParameterSample.h"
#include "Param/Distrib/RangedDistributions.h"
#include <utility>

namespace {

void checkIfEmpty(const std::vector<double>& input);

const std::string relative_resolution = "scanRelativeResolution";
const std::string absolute_resolution = "scanAbsoluteResolution";

class ScanSingleRelativeResolution : public IScanResolution {
public:
    ScanSingleRelativeResolution(const IRangedDistribution& distr, double reldev)
        : IScanResolution(distr)
        , m_reldev(reldev)
    {
    }
    ~ScanSingleRelativeResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanSingleRelativeResolution(*rdistribution(), m_reldev);
    }

    DistrOutput generateResolutionSamples(double mean, size_t n_times) const override;
    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return m_reldev; }

protected:
    std::string resolutionFactoryName() const override { return relative_resolution; }

private:
    double m_reldev; //!< deltas for computing resolutions
};

class ScanSingleAbsoluteResolution : public IScanResolution {
public:
    ScanSingleAbsoluteResolution(const IRangedDistribution& distr, double stddev)
        : IScanResolution(distr)
        , m_stddev(stddev)
    {
    }
    ~ScanSingleAbsoluteResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanSingleAbsoluteResolution(*rdistribution(), m_stddev);
    }

    DistrOutput generateResolutionSamples(double mean, size_t n_times) const override;
    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return m_stddev; }

protected:
    std::string resolutionFactoryName() const override { return absolute_resolution; }

private:
    double m_stddev; //!< deltas for computing resolutions
};

class ScanVectorRelativeResolution : public IScanResolution {
public:
    ScanVectorRelativeResolution(const IRangedDistribution& distr, std::vector<double> reldev)
        : IScanResolution(distr)
        , m_reldev(std::move(reldev))
    {
        checkIfEmpty(m_reldev);
    }
    ~ScanVectorRelativeResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanVectorRelativeResolution(*rdistribution(), m_reldev);
    }

    DistrOutput generateResolutionSamples(double mean, size_t n_times) const override;
    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return 0; } // TODO

protected:
    std::string resolutionFactoryName() const override { return relative_resolution; }

private:
    std::vector<double> m_reldev; //!< deltas for computing resolutions
};

class ScanVectorAbsoluteResolution : public IScanResolution {
public:
    ScanVectorAbsoluteResolution(const IRangedDistribution& distr, std::vector<double> stddev)
        : IScanResolution(distr)
        , m_stddev(std::move(stddev))
    {
        checkIfEmpty(m_stddev);
    }
    ~ScanVectorAbsoluteResolution() override = default;

    IScanResolution* clone() const override
    {
        return new ScanVectorAbsoluteResolution(*rdistribution(), m_stddev);
    }

    DistrOutput generateResolutionSamples(double mean, size_t n_times) const override;
    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return 0; } // TODO

protected:
    std::string resolutionFactoryName() const override { return absolute_resolution; }

private:
    std::vector<double> m_stddev; //!< deltas for computing resolutions
};

class ScanEmptyResolution : public IScanResolution {
public:
    ScanEmptyResolution() = default;

    ~ScanEmptyResolution() override = default;

    IScanResolution* clone() const override { return new ScanEmptyResolution(); }

    DistrOutput generateResolutionSamples(double mean, size_t n_times) const override;
    DistrOutput generateResolutionSamples(const std::vector<double>& mean) const override;
    std::vector<double> stdDevs(double mean, size_t n_times) const override;
    std::vector<double> stdDevs(const std::vector<double>& mean) const override;

    double delta() const override { return 0; }

protected:
    std::string resolutionFactoryName() const override;
};

} // namespace

//  ************************************************************************************************
//  IScanResolution implementation
//  ************************************************************************************************

IScanResolution::~IScanResolution() = default;

size_t IScanResolution::nSamples() const
{
    if (m_distr)
        return m_distr->nSamples();
    return 1;
}

IScanResolution::IScanResolution() = default;

IScanResolution::IScanResolution(const IRangedDistribution& distr)
    : m_distr(distr.clone())
{
}

//  ************************************************************************************************
//  child class implementations
//  ************************************************************************************************

namespace {

IScanResolution::DistrOutput
ScanSingleRelativeResolution::generateResolutionSamples(double mean, size_t n_times) const
{
    const double stddev = mean * m_reldev;
    return DistrOutput(n_times, rdistribution()->generateSamples(mean, stddev));
}

IScanResolution::DistrOutput
ScanSingleRelativeResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    checkIfEmpty(mean);
    DistrOutput result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back(rdistribution()->generateSamples(mean[i], mean[i] * m_reldev));
    return result;
}

std::vector<double> ScanSingleRelativeResolution::stdDevs(double mean, size_t n_times) const
{
    return std::vector<double>(n_times, mean * m_reldev);
}

std::vector<double> ScanSingleRelativeResolution::stdDevs(const std::vector<double>& mean) const
{
    checkIfEmpty(mean);
    std::vector<double> result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back(mean[i] * m_reldev);
    return result;
}

IScanResolution::DistrOutput
ScanVectorRelativeResolution::generateResolutionSamples(double mean, size_t n_times) const
{
    return generateResolutionSamples(std::vector<double>(n_times, mean));
}

IScanResolution::DistrOutput
ScanVectorRelativeResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    const std::vector<double> stddevs = stdDevs(mean);
    return rdistribution()->generateSamples(mean, stddevs);
}

std::vector<double> ScanVectorRelativeResolution::stdDevs(double mean, size_t n_times) const
{
    return stdDevs(std::vector<double>(n_times, mean));
}

std::vector<double> ScanVectorRelativeResolution::stdDevs(const std::vector<double>& mean) const
{
    const size_t result_size = mean.size();
    if (result_size != m_reldev.size())
        throw std::runtime_error(
            "Error in ScanVectorRelativeResolution::stdDevs: passed mean values vector "
            "size shall be of the same size with relative deviations vector");

    std::vector<double> stddevs(result_size);
    for (size_t i = 0; i < result_size; ++i)
        stddevs[i] = m_reldev[i] * mean[i];
    return stddevs;
}

IScanResolution::DistrOutput
ScanSingleAbsoluteResolution::generateResolutionSamples(double mean, size_t n_times) const
{
    return DistrOutput(n_times, rdistribution()->generateSamples(mean, m_stddev));
}

IScanResolution::DistrOutput
ScanSingleAbsoluteResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    checkIfEmpty(mean);
    DistrOutput result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back(rdistribution()->generateSamples(mean[i], m_stddev));
    return result;
}

std::vector<double> ScanSingleAbsoluteResolution::stdDevs(double, size_t n_times) const
{
    return std::vector<double>(n_times, m_stddev);
}

std::vector<double> ScanSingleAbsoluteResolution::stdDevs(const std::vector<double>& mean) const
{
    checkIfEmpty(mean);
    return std::vector<double>(mean.size(), m_stddev);
}

IScanResolution::DistrOutput
ScanVectorAbsoluteResolution::generateResolutionSamples(double mean, size_t n_times) const
{
    return generateResolutionSamples(std::vector<double>(n_times, mean));
}

IScanResolution::DistrOutput
ScanVectorAbsoluteResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    const size_t result_size = mean.size();
    if (result_size != m_stddev.size())
        throw std::runtime_error(
            "Error in ScanVectorAbsoluteResolution::generateResolutionSamples: passed mean values "
            "vector "
            "size shall be of the same size with standard deviations vector");
    return rdistribution()->generateSamples(mean, m_stddev);
}

std::vector<double> ScanVectorAbsoluteResolution::stdDevs(double mean, size_t n_times) const
{
    return stdDevs(std::vector<double>(n_times, mean));
}

std::vector<double> ScanVectorAbsoluteResolution::stdDevs(const std::vector<double>& mean) const
{
    const size_t result_size = mean.size();
    if (result_size != m_stddev.size())
        throw std::runtime_error(
            "Error in ScanVectorAbsoluteResolution::generateResolutionSamples: passed mean values "
            "vector "
            "size shall be of the same size with standard deviations vector");
    return m_stddev;
}

ScanEmptyResolution::DistrOutput
ScanEmptyResolution::generateResolutionSamples(double mean, size_t n_times) const
{
    return DistrOutput(n_times, std::vector<ParameterSample>{ParameterSample(mean, 1.)});
}

ScanEmptyResolution::DistrOutput
ScanEmptyResolution::generateResolutionSamples(const std::vector<double>& mean) const
{
    DistrOutput result;
    result.reserve(mean.size());
    for (size_t i = 0, size = mean.size(); i < size; ++i)
        result.push_back({ParameterSample(mean[i], 1.0)});
    return result;
}

std::vector<double> ScanEmptyResolution::stdDevs(double, size_t n_times) const
{
    return std::vector<double>(n_times, 0.0);
}

std::vector<double> ScanEmptyResolution::stdDevs(const std::vector<double>& mean) const
{
    return std::vector<double>(mean.size(), 0.0);
}

std::string ScanEmptyResolution::resolutionFactoryName() const
{
    throw std::runtime_error(
        "Error in ScanEmptyResolution::name: attempt to get a name of an empty resolution object.");
}

void checkIfEmpty(const std::vector<double>& input)
{
    if (input.empty())
        throw std::runtime_error("Error in IScanResolution: passed vector is empty");
}

} // namespace

//  ************************************************************************************************
//  Factory functions
//  ************************************************************************************************

IScanResolution* scanRelativeResolution(const IRangedDistribution& distr, double stddev)
{
    return new ScanSingleRelativeResolution(distr, stddev);
}

IScanResolution* scanRelativeVectorResolution(const IRangedDistribution& distr,
                                              const std::vector<double>& stddevs)
{
    return new ScanVectorRelativeResolution(distr, stddevs);
}

IScanResolution* scanAbsoluteResolution(const IRangedDistribution& distr, double stddev)
{
    return new ScanSingleAbsoluteResolution(distr, stddev);
}

IScanResolution* scanAbsoluteVectorResolution(const IRangedDistribution& distr,
                                              const std::vector<double>& stddevs)
{
    return new ScanVectorAbsoluteResolution(distr, stddevs);
}

IScanResolution* scanEmptyResolution()
{
    return new ScanEmptyResolution();
}
