//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Scan/ScanResolution.h
//! @brief     Defines scan resolution class.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SCAN_SCANRESOLUTION_H
#define BORNAGAIN_SIM_SCAN_SCANRESOLUTION_H

#include "Base/Types/ICloneable.h"
#include <memory>
#include <string>
#include <vector>

class IRangedDistribution;
class ParameterSample;

//! Abstract base class for reflectivity resolution functions.
//! Child classes have local scope only (declared and implemented in ScanResolution.cpp);
//! they can only be created through the factory functions declared below.
class IScanResolution : public ICloneable {
protected:
    using DistrOutput = std::vector<std::vector<ParameterSample>>;

public:
    ~IScanResolution() override;
#ifndef SWIG
    IScanResolution* clone() const override = 0;
    const IRangedDistribution* rdistribution() const
    {
        return m_distr.get();
    }
    size_t nSamples() const;

    virtual DistrOutput generateResolutionSamples(double mean, size_t n_times) const = 0;
    virtual DistrOutput generateResolutionSamples(const std::vector<double>& mean) const = 0;
    virtual std::vector<double> stdDevs(double mean, size_t n_times) const = 0;
    virtual std::vector<double> stdDevs(const std::vector<double>& mean) const = 0;

    virtual std::string resolutionFactoryName() const = 0;
    virtual double delta() const = 0;

#endif // SWIG
protected:
    IScanResolution();
    IScanResolution(const IRangedDistribution& distr);

private:
    std::unique_ptr<IRangedDistribution> m_distr; //!< basic distribution function
};

IScanResolution* scanRelativeResolution(const IRangedDistribution& distr, double stddev);
IScanResolution* scanRelativeVectorResolution(const IRangedDistribution& distr,
                                              const std::vector<double>& stddevs);
IScanResolution* scanAbsoluteResolution(const IRangedDistribution& distr, double stddev);
IScanResolution* scanAbsoluteVectorResolution(const IRangedDistribution& distr,
                                              const std::vector<double>& stddevs);
IScanResolution* scanEmptyResolution();

#endif // BORNAGAIN_SIM_SCAN_SCANRESOLUTION_H
