//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ISimulation.h
//! @brief     Defines interface ISimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_ISIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_ISIMULATION_H

#include "Param/Distrib/ParameterDistribution.h"
#include "Param/Node/INode.h"
#include <functional>
#include <heinz/Vectors3D.h>

class DistributionHandler;
class IBackground;
class IComputation;
class ICoordSystem;
class MultiLayer;
class ProgressHandler;
class ReSample;
class SimulationOptions;
class SimulationResult;

//! Abstract base class, holds the infrastructure to run a simulation.
//!
//! Base class of SpecularSimulation, OffspecSimulation, and DepthprobeSimulation.
//!
//! Holds an instrument and sample model. Provides the common infrastructure
//! to run a simulation: multithreading, batch processing, averaging over
//! parameter distributions, etc.
//!
//! Simulations are run, and results returned, by the function ISimulation::simulate().

class ISimulation : public INode {
public:
    ISimulation(const MultiLayer& sample);
    ~ISimulation() override;

    ISimulation(const ISimulation&) = delete;

    //... Setters:
    void setBackground(const IBackground& bg);

    void addParameterDistribution(ParameterDistribution::WhichParameter whichParameter,
                                  const IDistribution1D& distribution, size_t n_samples,
                                  double sigma_factor = 0.0,
                                  const RealLimits& limits = RealLimits());
    void addParameterDistribution(const ParameterDistribution& par_distr);

#ifndef SWIG
    void subscribe(const std::function<bool(size_t)>& inform);
#endif // SWIG

    void setTerminalProgressMonitor();

    SimulationOptions& options();

    //... Executor:
    SimulationResult simulate(); //!< Runs a simulation, and returns the result.

#ifndef SWIG
    //... Getters:
    std::vector<const INode*> nodeChildren() const override;

    const MultiLayer* sample() const;
    const IBackground* background() const;
    const std::vector<ParameterDistribution>& getDistributions() const;

    virtual const ICoordSystem* simCoordSystem() const = 0;

    const SimulationOptions& options() const;

protected:
    ProgressHandler& progress();

    DistributionHandler& distributionHandler();

    std::vector<double> m_cache;

private:
    //... Executor:
    void runSingleSimulation(const ReSample& re_sample, size_t batch_start, size_t batch_size,
                             double weight = 1.0);

    //... Virtual executors:
    virtual void initDistributionHandler() {}

    //! Put into a clean state for running a simulation.
    virtual void prepareSimulation() {}

    //! Initializes the vector of ISimulation elements.
    virtual void initElementVector() = 0;

    //! Generate a single threaded computation for a given range of simulation elements.
    //! @param re_sample Preprocessed version of our sample
    //! @param start Index of the first element to include into computation
    //! @param n_elements Number of elements to process
    virtual std::unique_ptr<IComputation> createComputation(const ReSample& re_sample, size_t start,
                                                            size_t n_elements) = 0;

    //! Normalize the detector counts to beam intensity, to solid angle, and to exposure angle.
    virtual void normalize(size_t start, size_t n_elements) = 0;

    virtual void addBackgroundIntensity(size_t start, size_t n_elements) = 0;

    virtual void addDataToCache(double weight) = 0;

    //... Virtual getters:
    //! Force polarized computation even in absence of sample magnetization or external fields
    virtual bool force_polarized() const = 0;

    //! Returns the number of elements this simulation needs to calculate.
    virtual size_t numberOfElements() const = 0;

    //! Returns the number of output channels to be computed. Determines size of m_cache.
    virtual size_t nOutChannels() const { return numberOfElements(); }

    //! Returns simulation result, based on intensity held in elements vector.
    virtual SimulationResult packResult() = 0;

    //! Checks the distribution validity for simulation.
    virtual void validateParametrization(const ParameterDistribution&) const {}

    //... Simulation model:
    std::unique_ptr<const MultiLayer> m_sample;
    std::unique_ptr<const IBackground> m_background;
    std::unique_ptr<SimulationOptions> m_options;

    //... Computational auxiliaries:
    std::unique_ptr<DistributionHandler> m_distribution_handler;
    std::unique_ptr<ProgressHandler> m_progress;
#endif // SWIG
};

#endif // BORNAGAIN_SIM_SIMULATION_ISIMULATION_H
