//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/OffspecSimulation.cpp
//! @brief     Implements class OffspecSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/OffspecSimulation.h"
#include "Base/Axis/Bin.h"
#include "Base/Axis/IAxis.h"
#include "Base/Pixel/IPixel.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Datafield.h"
#include "Device/Detector/IDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/DistributionHandler.h"
#include "Param/Distrib/Distributions.h"
#include "Resample/Element/DiffuseElement.h"
#include "Sim/Background/IBackground.h"
#include "Sim/Computation/DWBAComputation.h"

OffspecSimulation::OffspecSimulation(const Beam& beam, const MultiLayer& sample,
                                     const IDetector& detector)
    : ISimulation(sample)
    , m_beam(beam.clone())
    , m_detector(detector.clone())
{
    m_detector->setDetectorNormal(m_beam->ki());
}

OffspecSimulation::OffspecSimulation(const MultiLayer& sample)
    : ISimulation(sample)
    , m_beam(Beam::horizontalBeam().clone())
    , m_detector(new SphericalDetector())
{
}

OffspecSimulation::~OffspecSimulation() = default;

std::vector<const INode*> OffspecSimulation::nodeChildren() const
{
    std::vector<const INode*> result = ISimulation::nodeChildren();
    result.push_back(m_beam.get());
    if (m_detector)
        result.push_back(m_detector.get());
    return result;
}

void OffspecSimulation::prepareSimulation()
{
    ASSERT(m_detector->rank() == 2);
    ASSERT(m_alpha_i_axis);
    ASSERT(m_alpha_i_axis->size() >= 1);
    m_active_indices = m_detector->active_indices();
    m_pixels.reserve(m_active_indices.size());
    for (auto detector_index : m_active_indices)
        m_pixels.emplace_back(m_detector->createPixel(detector_index));
}

void OffspecSimulation::setBeamParameters(double wavelength, const IAxis& alpha_axis, double phi_i)
{
    m_alpha_i_axis.reset(alpha_axis.clone());
    if (alpha_axis.size() < 1)
        throw std::runtime_error("OffspecSimulation::prepareSimulation() "
                                 "-> Error. Incoming alpha range size < 1.");
    beam() = InBeam(beam().intensity(), wavelength, alpha_axis.min(), phi_i);
}

const IAxis* OffspecSimulation::beamAxis() const
{
    return m_alpha_i_axis.get();
}

const ICoordSystem* OffspecSimulation::simCoordSystem() const
{
    ASSERT(beamAxis());
    return m_detector->offspecCoords(beamAxis()->clone(), beam().ki());
}

void OffspecSimulation::initElementVector()
{
    m_eles.clear();
    Beam beam2 = beam();
    for (size_t i = 0; i < m_alpha_i_axis->size(); ++i) {
        beam2.setInclination(m_alpha_i_axis->bin(i).center());
        std::vector<DiffuseElement> eles_i = generateElements(beam2);
        for (auto&& ele : eles_i)
            m_eles.emplace_back(std::move(ele));
    }
    if (m_cache.empty())
        m_cache.resize(m_eles.size(), 0.0);
}

void OffspecSimulation::validateParametrization(const ParameterDistribution& par_distr) const
{
    const bool zero_mean = par_distr.getDistribution()->mean() == 0.0;
    if (zero_mean)
        return;
    /* TODO come back to this after refactoring of distribution handling
    if (par_distr.whichParameter() == ParameterDistribution::BeamInclinationAngle)
        throw std::runtime_error("Error in OffspecSimulation: parameter distribution of "
                                 "beam inclination angle should have zero mean.");
    */
}

//... Overridden executors:

//! init callbacks for setting the parameter values
void OffspecSimulation::initDistributionHandler()
{
    for (const auto& distribution : distributionHandler().getDistributions()) {
        const auto* dc = const_cast<ParameterDistribution*>(&distribution);

        switch (dc->whichParameter()) {
        case ParameterDistribution::BeamAzimuthalAngle:
            distributionHandler().defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setAzimuthalAngle(d); });
            break;
        case ParameterDistribution::BeamInclinationAngle:
            distributionHandler().defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setInclination(d); });
            break;
        case ParameterDistribution::BeamWavelength:
            distributionHandler().defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setWavelength(d); });
            break;
        default:
            ASSERT(0);
        }
    }
}

std::unique_ptr<IComputation> OffspecSimulation::createComputation(const ReSample& re_sample,
                                                                   size_t start, size_t n_elements)
{
    ASSERT(start < m_eles.size() && start + n_elements <= m_eles.size());
    const auto& begin = m_eles.begin() + static_cast<long>(start);
    return std::make_unique<DWBAComputation>(re_sample, options(), progress(), begin,
                                             begin + static_cast<long>(n_elements));
}

void OffspecSimulation::normalize(size_t start, size_t n_elements)
{
    const double beam_intensity = m_beam->intensity();
    for (size_t i = start; i < start + n_elements; ++i) {
        DiffuseElement& element = m_eles[i];
        double sin_alpha_i = std::abs(std::sin(element.alphaI()));
        if (sin_alpha_i == 0.0) {
            element.setIntensity(0);
            continue;
        }
        const double solid_angle = element.solidAngle();
        element.setIntensity(element.intensity() * beam_intensity * solid_angle / sin_alpha_i);
    }
}

void OffspecSimulation::addBackgroundIntensity(size_t start, size_t n_elements)
{
    if (!background())
        return;
    for (size_t i = start, stop_point = start + n_elements; i < stop_point; ++i) {
        DiffuseElement& element = m_eles[i];
        element.setIntensity(background()->addBackground(element.intensity()));
    }
}

void OffspecSimulation::addDataToCache(double weight)
{
    ASSERT(m_eles.size() == m_cache.size());
    for (size_t i = 0; i < m_eles.size(); i++)
        m_cache[i] += m_eles[i].intensity() * weight;
}

//... Overridden getters:

bool OffspecSimulation::force_polarized() const
{
    return m_detector->analyzer().analyzerDirection() != R3{};
}

size_t OffspecSimulation::numberOfElements() const
{
    return m_active_indices.size() * m_alpha_i_axis->size();
}

SimulationResult OffspecSimulation::packResult()
{

    ASSERT(m_eles.size() == m_cache.size());
    for (size_t i = 0; i < m_eles.size(); i++)
        m_eles[i].setIntensity(m_cache[i]);

    // update intensity map
    ASSERT(m_alpha_i_axis);
    Datafield intensity_map({m_alpha_i_axis->clone(), m_detector->axis(1).clone()});
    intensity_map.setAllTo(0.);

    const IAxis& phi_axis = m_detector->axis(0);
    size_t phi_f_size = phi_axis.size();
    ASSERT(phi_f_size * intensity_map.size() == m_eles.size());

    size_t ny = m_detector->axis(1).size();

    // Normalize, apply detector resolution and transfer detector image corresponding to
    // alpha_i = m_alpha_i_axis->bin(index)
    for (size_t j = 0; j < m_alpha_i_axis->size(); ++j) {
        Datafield detector_image({m_detector->axis(0).clone(), m_detector->axis(1).clone()});
        size_t N = detector_image.size();
        for (size_t i = 0; i < N; ++i)
            detector_image[i] = m_eles[j * N + i].intensity();
        m_detector->applyDetectorResolution(&detector_image);
        for (size_t i = 0; i < N; ++i)
            intensity_map[j * ny + i % ny] += detector_image[i];
    }

    return {intensity_map, simCoordSystem()};
}

//... Local functions:

std::vector<DiffuseElement> OffspecSimulation::generateElements(const Beam& beam)
{
    const double wavelength = beam.wavelength();
    const double alpha_i = beam.alpha_i();
    const double phi_i = beam.phi_i();
    const SpinMatrix beam_polMatrices = beam.polMatrix();
    const SpinMatrix analyzer_operator = m_detector->analyzer().matrix();
    const size_t i_specular = m_detector->indexOfSpecular(beam);
    const size_t N = m_active_indices.size();

    std::vector<DiffuseElement> result;
    result.reserve(N);
    for (size_t i = 0; i < N; ++i)
        result.emplace_back(DiffuseElement(
            wavelength, alpha_i, phi_i, std::unique_ptr<IPixel>(m_pixels[i]->clone()),
            beam_polMatrices, analyzer_operator, m_active_indices[i] == i_specular));
    return result;
}
