//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ScatteringSimulation.cpp
//! @brief     Implements interface ISimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/ScatteringSimulation.h"
#include "Base/Axis/IAxis.h"
#include "Base/Pixel/IPixel.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Data/Datafield.h"
#include "Device/Detector/IDetector.h"
#include "Device/Detector/SimulationAreaIterator.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/DistributionHandler.h"
#include "Resample/Element/DiffuseElement.h"
#include "Sim/Background/IBackground.h"
#include "Sim/Computation/DWBAComputation.h"

namespace {

const Datafield* createDetectorIntensity(const IDetector& detector,
                                         const std::vector<DiffuseElement>& elements)
{
    std::unique_ptr<Datafield> detectorMap(detector.createDetectorMap());
    ASSERT(detectorMap);
    size_t elementIndex = 0;
    for (auto it = detector.beginNonMaskedPoints(); it != detector.endNonMaskedPoints(); ++it)
        (*detectorMap)[it.roiIndex()] = elements[elementIndex++].intensity();

    detector.applyDetectorResolution(detectorMap.get());

    return detectorMap.release();
}

} // namespace

ScatteringSimulation::ScatteringSimulation(const Beam& beam, const MultiLayer& sample,
                                           const IDetector& detector)
    : ISimulation(sample)
    , m_beam(beam.clone())
    , m_detector(detector.clone())
{
    m_detector->setDetectorNormal(m_beam->ki());
    ASSERT(m_detector->rank() == 2);
}

ScatteringSimulation::~ScatteringSimulation() = default;

void ScatteringSimulation::addMask(const IShape2D& shape, bool mask_value)
{
    m_detector->addMask(shape, mask_value);
}

void ScatteringSimulation::maskAll()
{
    m_detector->maskAll();
}

void ScatteringSimulation::setRegionOfInterest(double xlow, double ylow, double xup, double yup)
{
    m_detector->setRegionOfInterest(xlow, ylow, xup, yup);
}

const ICoordSystem* ScatteringSimulation::simCoordSystem() const
{
    auto* const det2D = dynamic_cast<const IDetector*>(getDetector());
    ASSERT(det2D);
    return det2D->scatteringCoords(beam());
}

//... Overridden executors:

//! init callbacks for setting the parameter values
void ScatteringSimulation::initDistributionHandler()
{
    for (const auto& distribution : distributionHandler().getDistributions()) {
        const auto* dc = const_cast<ParameterDistribution*>(&distribution);

        switch (dc->whichParameter()) {
        case ParameterDistribution::BeamAzimuthalAngle:
            distributionHandler().defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setAzimuthalAngle(d); });
            break;
        case ParameterDistribution::BeamInclinationAngle:
            distributionHandler().defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setInclination(d); });
            break;
        case ParameterDistribution::BeamWavelength:
            distributionHandler().defineCallbackForDistribution(
                dc, [&](double d) { m_beam->setWavelength(d); });
            break;
        default:
            ASSERT(0);
        }
    }
}

void ScatteringSimulation::prepareSimulation()
{
    m_active_indices = m_detector->active_indices();
    m_pixels.reserve(m_active_indices.size());
    for (auto detector_index : m_active_indices)
        m_pixels.emplace_back(m_detector->createPixel(detector_index));
}

void ScatteringSimulation::initElementVector()
{
    m_eles = generateElements(beam());
}

std::unique_ptr<IComputation>
ScatteringSimulation::createComputation(const ReSample& re_sample, size_t start, size_t n_elements)
{
    ASSERT(start < m_eles.size() && start + n_elements <= m_eles.size());
    const auto& begin = m_eles.begin() + static_cast<long>(start);
    return std::make_unique<DWBAComputation>(re_sample, options(), progress(), begin,
                                             begin + static_cast<long>(n_elements));
}

void ScatteringSimulation::normalize(size_t start, size_t n_elements)
{
    for (size_t i = start; i < start + n_elements; ++i) {
        DiffuseElement& element = m_eles[i];
        double sin_alpha_i = std::abs(std::sin(element.alphaI()));
        if (sin_alpha_i == 0.0) {
            element.setIntensity(0);
            continue;
        }
        const double solid_angle = element.solidAngle();
        element.setIntensity(element.intensity() * m_beam->intensity() * solid_angle / sin_alpha_i);
    }
}

void ScatteringSimulation::addBackgroundIntensity(size_t start, size_t n_elements)
{
    if (!background())
        return;
    for (size_t i = start, stop_point = start + n_elements; i < stop_point; ++i) {
        DiffuseElement& element = m_eles[i];
        element.setIntensity(background()->addBackground(element.intensity()));
    }
}

void ScatteringSimulation::addDataToCache(double weight)
{
    ASSERT(m_eles.size() == m_cache.size());
    for (size_t i = 0; i < m_eles.size(); i++)
        m_cache[i] += m_eles[i].intensity() * weight;
}

//... Overridden getters:

bool ScatteringSimulation::force_polarized() const
{
    return m_detector->analyzer().analyzerDirection() != R3{};
}

size_t ScatteringSimulation::numberOfElements() const
{
    return m_active_indices.size();
}

SimulationResult ScatteringSimulation::packResult()
{
    ASSERT(m_eles.size() == m_cache.size());
    for (size_t i = 0; i < m_eles.size(); i++)
        m_eles[i].setIntensity(m_cache[i]);

    auto* const coordsys = simCoordSystem();
    std::unique_ptr<const Datafield> data(createDetectorIntensity(*m_detector, m_eles));
    return {*data, *coordsys};
}

std::vector<DiffuseElement> ScatteringSimulation::generateElements(const Beam& beam) const
{
    const size_t i_specular = m_detector->indexOfSpecular(beam);
    const size_t N = m_active_indices.size();

    std::vector<DiffuseElement> result;
    result.reserve(N);
    for (size_t i = 0; i < N; ++i)
        result.emplace_back(DiffuseElement(beam.wavelength(), beam.alpha_i(), beam.phi_i(),
                                           std::unique_ptr<IPixel>(m_pixels[i]->clone()),
                                           beam.polMatrix(), m_detector->analyzer().matrix(),
                                           m_active_indices[i] == i_specular));
    return result;
}
