//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/SpecularSimulation.h
//! @brief     Defines class SpecularSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_SPECULARSIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_SPECULARSIMULATION_H

#include "Sim/Simulation/ISimulation.h"

class IAxis;
class IFootprintFactor;
class ISpecularScan;
class SpecularElement;

//! Specular reflectometry simulation.
//!
//! Holds an instrument and sample model.
//! Computes detected intensity as function of the glancing angles alpha_i=alpha_f.
//! @ingroup simulation

class SpecularSimulation : public ISimulation {
public:
    SpecularSimulation(const ISpecularScan& scan, const MultiLayer& sample);
    ~SpecularSimulation() override;

    std::string className() const final { return "SpecularSimulation"; }

#ifndef SWIG
    const ICoordSystem* simCoordSystem() const override;
    //! Returns internal data handler
    const ISpecularScan* scan() const
    {
        return m_scan.get();
    }

private:
    //... Overridden executors:
    void initElementVector() override;

    //! Generates a single threaded computation for a given range of simulation elements
    std::unique_ptr<IComputation> createComputation(const ReSample& re_sample, size_t start,
                                                    size_t n_elements) override;

    //! Normalizes the detector counts to beam intensity, to solid angle, and to exposure angle.
    void normalize(size_t start_ind, size_t n_elements) override;

    void addBackgroundIntensity(size_t start_ind, size_t n_elements) override;

    void addDataToCache(double weight) override;

    //... Overridden getters:
    bool force_polarized() const override;

    //! Returns the number of elements this simulation needs to calculate
    size_t numberOfElements() const override;

    SimulationResult packResult() override;

    //! Checks the distribution validity for simulation.
    void validateParametrization(const ParameterDistribution& par_distr) const override;

    //... Model components:
    std::unique_ptr<const ISpecularScan> m_scan;

    //... Caches:
    std::vector<SpecularElement> m_eles;
#endif // SWIG
};

#endif // BORNAGAIN_SIM_SIMULATION_SPECULARSIMULATION_H
