//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Tests/Suite/Common/TestSuite.h
//! @brief     Provides TEST_F stances that define the BornAgain standard tests.
//!
//!            Each of these stances runs one standard test.
//!            The function run, defined in file RunTest.cpp, is the same for Core/Py/GUI std tests.
//!            It calls a function checkSimulation that is different for those three test suites.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_TESTS_SUITE_COMMON_TESTSUITE_H
#define BORNAGAIN_TESTS_SUITE_COMMON_TESTSUITE_H

#include "Base/Axis/IAxis.h"
#include "Sample/ComponentBuilder/FormFactorComponents.h"
#include "Sample/ComponentBuilder/Profile2DComponents.h"
#include "Sample/Multilayer/MultiLayer.h"
#include "Sample/StandardSamples/ExemplarySamples.h"
#include "Sim/Background/ConstantBackground.h"
#include "Sim/Simulation/includeSimulations.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/SimFactory/MakeSimulations.h"

// By defining DISABLE_DISTRIBUTION_TESTS in the project, tests which contain particle distributions
// can be set do DISABLED. This is useful since the tests in here are the same for Core/Py/GUI, but
// the particle distributions have been removed from GUI.
#ifdef DISABLE_DISTRIBUTION_TESTS
#define DISTRIBUTION_TEST(test_fixture, name) TEST(test_fixture, DISABLED_##name)
#else
#define DISTRIBUTION_TEST(test_fixture, name) TEST(test_fixture, name)
#endif

TEST(TESTNAME, FormFactors)
{
    for (const std::string& ffname : FormFactorComponents().keys()) {
        const IFormFactor* ff = FormFactorComponents().getItem(ffname)->clone();
        ASSERT(ff);
        std::unique_ptr<const MultiLayer> sample(
            ExemplarySamples::createParticleInVacuumWithFF(ff));
        ASSERT(sample);
        auto sim = test::makeSimulation::MiniGISAS(*sample);
        EXPECT_TRUE(runTest("FormFactors_" + ff->className(), *sim, 2e-10));
    }
}

TEST(TESTNAME, GISASAbsorptiveSLDLayers)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createLayersWithAbsorptionBySLD());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("GISASAbsorptiveSLDLayers", *sim, 2e-10));
}

TEST(TESTNAME, CylindersAndPrisms)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersAndPrisms());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CylindersAndPrisms", *sim, 2e-10));
}

TEST(TESTNAME, RadialParacrystal)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createRadialParacrystal());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("RadialParacrystal", *sim, 2e-10));
}

TEST(TESTNAME, HardDisk)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createHardDisk());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("HardDisk", *sim, 2e-10));
}

TEST(TESTNAME, Basic2DParacrystal)
{
    for (const std::string& pdfname : Profile2DComponents().keys()) {
        const IProfile2D* pdf2 = Profile2DComponents().getItem(pdfname)->clone();
        ASSERT(pdf2);
        std::cout << pdf2->pythonConstructor() << std::endl;
        std::unique_ptr<const MultiLayer> sample(
            ExemplarySamples::createBasic2DParacrystalWithFTDis(pdf2));
        ASSERT(sample);
        auto sim = test::makeSimulation::MiniGISAS(*sample);
        EXPECT_TRUE(runTest("Basic2DParacrystal_" + pdf2->className(), *sim, 2e-10));
    }
}

TEST(TESTNAME, HexParacrystal)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createHexParacrystal());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("HexParacrystal", *sim, 2e-10));
}

TEST(TESTNAME, Lattice1D)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createLattice1D());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("Lattice1D", *sim, 2e-10));
}

TEST(TESTNAME, RectParacrystal)
{
    // TODO: investigate numeric integration, which has become problematic under Windows
    //       after moving code from Core/Tools to Base/Utils
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createRectParacrystal());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("RectParacrystal", *sim, 1e-9));
}

TEST(TESTNAME, CoreShellParticle)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCoreShellParticle());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CoreShellParticle", *sim, 2e-10));
}

TEST(TESTNAME, CoreShellBoxRotateZandY)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCoreShellBoxRotateZandY());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CoreShellBoxRotateZandY", *sim, 2e-10));
}

TEST(TESTNAME, MultiLayerWithRoughness)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMultiLayerWithRoughness());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("MultiLayerWithRoughness", *sim, 2e-10));
}

TEST(TESTNAME, SquareLattice2D)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSquareLattice2D());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("SquareLattice2D", *sim, 2e-10));
}

TEST(TESTNAME, CenteredSquareLattice2D)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCenteredSquareLattice2D());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CenteredSquareLattice2D", *sim, 2e-10));
}

TEST(TESTNAME, RotatedSquareLattice2D)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createRotatedSquareLattice2D());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("RotatedSquareLattice2D", *sim, 2e-10));
}

TEST(TESTNAME, FiniteSquareLattice2D)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFiniteSquareLattice2D());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("FiniteSquareLattice2D", *sim, 2e-10));
}

TEST(TESTNAME, RotatedPyramids)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createRotatedPyramids());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("RotatedPyramids", *sim, 2e-10));
}

TEST(TESTNAME, ThickAbsorptiveSampleWithRoughness)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createThickAbsorptiveSample());
    auto sim = test::makeSimulation::ExtraLongWavelengthGISAS(*sample);
    EXPECT_TRUE(runTest("ThickAbsorptiveSampleWithRoughness", *sim, 2e-10));
}

TEST(TESTNAME, Compound)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCompound());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("Compound", *sim, 2e-10));
}

TEST(TESTNAME, BoxCompositionRotateX)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createBoxCompositionRotateX());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("BoxCompositionRotateX", *sim, 2e-10));
}

TEST(TESTNAME, BoxCompositionRotateY)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createBoxCompositionRotateY());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("BoxCompositionRotateY", *sim, 2e-10));
}

TEST(TESTNAME, BoxCompositionRotateZ)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createBoxCompositionRotateZ());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("BoxCompositionRotateZ", *sim, 2e-10));
}

TEST(TESTNAME, BoxCompositionRotateZandY)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createBoxCompositionRotateZandY());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("BoxCompositionRotateZandY", *sim, 2e-10));
}

TEST(TESTNAME, BoxStackComposition)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createBoxStackComposition());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("BoxStackComposition", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, CylindersWithSizeDistribution)
{
    std::unique_ptr<const MultiLayer> sample(
        ExemplarySamples::createCylindersWithSizeDistribution());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CylindersWithSizeDistribution", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, TwoTypesCylindersDistribution)
{
    std::unique_ptr<const MultiLayer> sample(
        ExemplarySamples::createTwoTypesCylindersDistribution());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("TwoTypesCylindersDistribution", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, RotatedPyramidsDistribution)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createRotatedPyramidsDistribution());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("RotatedPyramidsDistribution", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, SpheresWithLimitsDistribution)
{
    std::unique_ptr<const MultiLayer> sample(
        ExemplarySamples::createSpheresWithLimitsDistribution());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("SpheresWithLimitsDistribution", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, ConesWithLimitsDistribution)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createConesWithLimitsDistribution());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("ConesWithLimitsDistribution", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, LinkedBoxDistribution)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createLinkedBoxDistribution());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("LinkedBoxDistribution", *sim, 2e-10));
}

TEST(TESTNAME, MultipleLayout)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMultipleLayout());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("MultipleLayout", *sim, 2e-10));
}

TEST(TESTNAME, ApproximationDA)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSizeDistributionDAModel());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("ApproximationDA", *sim, 2e-10));
}

TEST(TESTNAME, ApproximationLMA)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSizeDistributionLMAModel());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("ApproximationLMA", *sim, 2e-10));
}

TEST(TESTNAME, ApproximationSSCA)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSizeDistributionSSCAModel());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("ApproximationSSCA", *sim, 2e-10));
}

DISTRIBUTION_TEST(TESTNAME, CylindersInSSCA)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInSSCA());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CylindersInSSCA", *sim, 2e-10));
}

TEST(TESTNAME, CosineRipple)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCosineRipple());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CosineRipple", *sim, 2e-10));
}

TEST(TESTNAME, TriangularRipple)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createTriangularRipple());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("TriangularRipple", *sim, 2e-10));
}

TEST(TESTNAME, AsymRipple)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createAsymRipple());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("AsymRipple", *sim, 2e-10));
}

TEST(TESTNAME, Mesocrystal)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMesocrystal());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("Mesocrystal", *sim, 2e-10));
}

TEST(TESTNAME, CustomMorphology)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCustomMorphology());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("CustomMorphology", *sim, 2e-10));
}

TEST(TESTNAME, TransformBox)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createTransformBox());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("TransformBox", *sim, 1e-10));
}

TEST(TESTNAME, MagneticParticleZeroField)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMagneticParticleZeroField());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("MagneticParticleZeroField", *sim, 2e-10));
}

TEST(TESTNAME, SlicedComposition)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSlicedComposition());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("SlicedComposition", *sim, 1e-10));
}

TEST(TESTNAME, MagneticSubstrateZeroField)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMagneticSubstrateZeroField());
    auto sim = test::makeSimulation::MiniZPolarizedGISAS(*sample, "PP");
    EXPECT_TRUE(runTest("MagneticSubstrateZeroField", *sim, 2e-10));
}

TEST(TESTNAME, MagneticRotation)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMagneticRotation());
    auto sim = test::makeSimulation::MiniZPolarizedGISAS(*sample, "PM");
    EXPECT_TRUE(runTest("MagneticRotationZPM", *sim, 2e-10));
}

TEST(TESTNAME, MagneticSpheres)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMagneticSpheres());
    auto sim = test::makeSimulation::MiniZPolarizedGISAS(*sample, "PM");
    EXPECT_TRUE(runTest("MagneticSpheres", *sim, 2e-10));
}

TEST(TESTNAME, MagneticCylinders)
{
    for (const std::string polCase : {"PP", "MP", "PM", "MM"}) {
        std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMagneticCylinders());
        auto sim = test::makeSimulation::MiniZPolarizedGISAS(*sample, polCase);
        EXPECT_TRUE(runTest("MagneticCylinders" + polCase, *sim, 2e-10));
    }
}

TEST(TESTNAME, MagneticSpheresInMagLayer)
{
    for (const std::string polCase : {"PP", "MP"}) {
        std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createMagneticLayer());
        auto sim = test::makeSimulation::MiniZPolarizedGISAS(*sample, polCase);
        EXPECT_TRUE(runTest("MagneticSpheresInMagLayer" + polCase, *sim, 2e-10));
    }
}

TEST(TESTNAME, BeamDivergence)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInBA());
    auto sim = test::makeSimulation::MiniGISASBeamDivergence(*sample);
    EXPECT_TRUE(runTest("BeamDivergence", *sim, 2e-10));
}

TEST(TESTNAME, DetectorResolution)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInBA());
    auto sim = test::makeSimulation::MiniGISASDetectorResolution(*sample);
    EXPECT_TRUE(runTest("DetectorResolution", *sim, 2e-10));
}

TEST(TESTNAME, SimulationWithMasks)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersAndPrisms());
    auto sim = test::makeSimulation::GISASWithMasks(*sample);
    EXPECT_TRUE(runTest("SimulationWithMasks", *sim, 1e-10));
}

TEST(TESTNAME, RectDetectorGeneric)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInBA());
    auto sim = test::makeSimulation::RectDetectorGeneric(*sample);
    EXPECT_TRUE(runTest("RectDetectorGeneric", *sim, 1e-10));
}

TEST(TESTNAME, RectDetectorPerpToSample)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInBA());
    auto sim = test::makeSimulation::RectDetectorPerpToSample(*sample);
    EXPECT_TRUE(runTest("RectDetectorPerpToSample", *sim, 1e-10));
}

TEST(TESTNAME, RectDetectorPerpToDirectBeam)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInBA());
    auto sim = test::makeSimulation::RectDetectorPerpToDirectBeam(*sample);
    EXPECT_TRUE(runTest("RectDetectorPerpToDirectBeam", *sim, 1e-10));
}

TEST(TESTNAME, RectDetectorPerpToReflectedBeam)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersInBA());
    auto sim = test::makeSimulation::RectDetectorPerpToReflectedBeam(*sample);
    EXPECT_TRUE(runTest("RectDetectorPerpToReflectedBeam", *sim, 1e-10));
}

TEST(TESTNAME, LargeCylindersMonteCarlo)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createLargeCylindersInDWBA());
    auto sim = test::makeSimulation::MiniGISASMonteCarlo(*sample);
    EXPECT_TRUE(runTest("LargeCylindersMonteCarlo", *sim, 5e-1));
}

TEST(TESTNAME, SphericalDetWithRoi)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersAndPrisms());
    auto sim = test::makeSimulation::SphericalDetWithRoi(*sample);
    EXPECT_TRUE(runTest("SphericalDetWithRoi", *sim, 1e-10));
}

TEST(TESTNAME, RectDetWithRoi)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createCylindersAndPrisms());
    auto sim = test::makeSimulation::RectDetWithRoi(*sample);
    EXPECT_TRUE(runTest("RectDetWithRoi", *sim, 1e-10));
}

TEST(TESTNAME, BoxesWithSpecular)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createBoxesSquareLattice2D());
    auto sim = test::makeSimulation::MiniGISASSpecularPeak(*sample);
    EXPECT_TRUE(runTest("BoxesWithSpecular", *sim, 1e-10));
}

TEST(TESTNAME, ConstantBackground)
{
    auto* sample = ExemplarySamples::createCylindersInBA();
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    sim->setBackground(ConstantBackground(1e3));
    EXPECT_TRUE(runTest("ConstantBackground", *sim, 1e-10));
}

TEST(TESTNAME, HomogeneousTiNiSample)
{
    auto* sample = ExemplarySamples::createHomogeneousMultilayer();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("HomogeneousTiNiSample", *sim, 1e-10));
}

TEST(TESTNAME, HomogeneousTiNiSampleWithAbsorption)
{
    auto* sample = ExemplarySamples::createPlainMultiLayerBySLD();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("HomogeneousTiNiSampleWithAbsorption", *sim, 1e-10));
}

TEST(TESTNAME, RoughnessInSpecular)
{
    auto* sample = ExemplarySamples::createMultiLayerWithRoughness();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("RoughnessInSpecular", *sim, 2e-9));
}

TEST(TESTNAME, GaussianBeamFootprint)
{
    auto* sample = ExemplarySamples::createHomogeneousMultilayer();
    auto sim = test::makeSimulation::SpecularWithGaussianBeam(*sample);
    EXPECT_TRUE(runTest("GaussianBeamFootprint", *sim, 1e-10));
}

TEST(TESTNAME, SquareBeamFootprint)
{
    auto* sample = ExemplarySamples::createHomogeneousMultilayer();
    auto sim = test::makeSimulation::SpecularWithSquareBeam(*sample);
    EXPECT_TRUE(runTest("SquareBeamFootprint", *sim, 1e-10));
}

TEST(TESTNAME, SpecularDivergentBeam)
{
    auto* sample = ExemplarySamples::createHomogeneousMultilayer();
    auto sim = test::makeSimulation::SpecularDivergentBeam(*sample);
    EXPECT_TRUE(runTest("SpecularDivergentBeam", *sim, 1e-10));
}

//  ************************************************************************************************
//  TODO: broken under GUI
//  ************************************************************************************************

#ifndef GUI_STD_TEST

TEST(TESTNAME, RelativeResolutionTOF)
{
    auto* sample = ExemplarySamples::createPlainMultiLayerBySLD();
    auto sim = test::makeSimulation::TOFRWithRelativeResolution(*sample);
    EXPECT_TRUE(runTest("RelativeResolutionTOF", *sim, 1e-10));
}

#endif // GUI_STD_TEST

//  ************************************************************************************************
//  TODO: broken under Python
//  ************************************************************************************************

#ifndef PYTHON_STD_TEST

TEST(TESTNAME, OffspecResonator)
{
    auto* sample = ExemplarySamples::createResonator();
    auto sim = test::makeSimulation::MiniOffspec(*sample);
    EXPECT_TRUE(runTest("OffspecResonator", *sim, 1e-10));
}

TEST(TESTNAME, FormFactorsWithAbsorption)
{
    for (const std::string& ffname : FormFactorComponents().keys()) {
        const IFormFactor* ff = FormFactorComponents().getItem(ffname)->clone();
        ASSERT(ff);
        std::unique_ptr<const MultiLayer> sample(
            ExemplarySamples::createLayersWithAbsorptionWithFF(ff));
        ASSERT(sample);
        auto sim = test::makeSimulation::MiniGISAS_v2(*sample);
        EXPECT_TRUE(runTest("FormFactorsWithAbsorption_" + ff->className(), *sim, 2e-10));
    }
}
#endif // PYTHON_STD_TEST

//  ************************************************************************************************
//  TODO: broken under GUI and Python
//  ************************************************************************************************

#ifndef PYTHON_STD_TEST
#ifndef GUI_STD_TEST

TEST(TESTNAME, NCRoughnessInSpecular)
{
    auto* sample = ExemplarySamples::createMultiLayerWithNCRoughness();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("NCRoughnessInSpecular", *sim, 2e-9));
}

TEST(TESTNAME, SuperLattice)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSuperLattice());
    auto sim = test::makeSimulation::MiniGISAS(*sample);
    EXPECT_TRUE(runTest("SuperLattice", *sim, 2e-10));
}

TEST(TESTNAME, SpecularWithSlicing1)
{
    // data almost fully agree with SpecularWithSlicing2, as verified by ConsistenceTest
    auto* sample = ExemplarySamples::createSlicedCylinders();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("SpecularWithSlicing1", *sim, 1e-10));
}

TEST(TESTNAME, SpecularWithSlicing2)
{
    auto* sample = ExemplarySamples::createSLDSlicedCylinders();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("SpecularWithSlicing2", *sim, 1e-10));
    auto simQ = test::makeSimulation::BasicSpecular(*sample, true);
    EXPECT_TRUE(runTest("SpecularWithSlicing2Q", *simQ, 1e-10));
}

TEST(TESTNAME, SpecularWithSlicing3)
{
    // data fully agree with SpecularWithSlicing2, as verified by ConsistenceTest
    auto* sample = ExemplarySamples::createAveragedSlicedCylinders();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("SpecularWithSlicing3", *sim, 1e-10));
}

TEST(TESTNAME, InstrumentDefinitionComparison)
{
    auto* sample = ExemplarySamples::createPlainMultiLayerBySLD();
    auto sim = test::makeSimulation::BasicSpecular(*sample, false);
    EXPECT_TRUE(runTest("InstrumentDefinitionComparison_0", *sim, 1e-10));
    auto simQ = test::makeSimulation::BasicSpecular(*sample, true);
    EXPECT_TRUE(runTest("InstrumentDefinitionComparison_Q", *simQ, 1e-10));
}

TEST(TESTNAME, TOFResolutionComparison)
{
    auto* sample = ExemplarySamples::createPlainMultiLayerBySLD();
    auto sim1 = test::makeSimulation::TOFRWithRelativeResolution(*sample);
    EXPECT_TRUE(runTest("TOFResolutionComparison_TR", *sim1, 1e-10));
    auto sim2 = test::makeSimulation::TOFRWithPointwiseResolution(*sample);
    EXPECT_TRUE(runTest("TOFResolutionComparison_TP", *sim2, 1e-10));
}

TEST(TESTNAME, BasicSpecular)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSimpleMagneticLayer());
    for (const std::string polCase : {"PP", "MM"}) {
        auto sim = test::makeSimulation::BasicYPolarizedSpecular(*sample, polCase, false);
        EXPECT_TRUE(runTest("BasicSpecular" + polCase, *sim, 1e-10));
    }
}

TEST(TESTNAME, PolarizedQAngleReflectivity)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createSimpleMagneticLayer());
    for (const std::string polKey : {"PP", "MM"}) {
        auto sim = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedQAngleReflectivity" + polKey + "_Q", *sim, 1e-10));
    }
}

TEST(TESTNAME, MagneticRotationReflectivity)
{
    for (const auto roughnessKey : {"Flat", "Tanh", "NevotCroce"}) {
        std::unique_ptr<const MultiLayer> sample(
            ExemplarySamples::createSimpleMagneticRotationWithRoughness(roughnessKey));
        for (const std::string polKey : {"PP", "PM", "MP", "MM"}) {
            auto sim = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
            EXPECT_TRUE(
                runTest("MagneticRotationReflectivity" + polKey + "_" + roughnessKey, *sim, 1e-10));
        }
    }
}

TEST(TESTNAME, PolarizedFeNiBilayer)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFeNiBilayer());
    for (const std::string polKey : {"PP", "MM"}) {
        auto sim1 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayer" + polKey, *sim1, 1e-7));
        auto sim2 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayer" + polKey + "_Q", *sim2, 1e-7));
    }
}

TEST(TESTNAME, PolarizedFeNiBilayerTanh)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFeNiBilayerTanh());
    for (const std::string polKey : {"PP", "MM"}) {
        auto sim1 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerTanh" + polKey, *sim1, 1e-7));
        auto sim2 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerTanh" + polKey + "_Q", *sim2, 1e-7));
    }
}

TEST(TESTNAME, PolarizedFeNiBilayerNC)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFeNiBilayerNC());
    for (const std::string polKey : {"PP", "MM"}) {
        auto sim1 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerNC" + polKey, *sim1, 1e-7));
        auto sim2 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerNC" + polKey + "_Q", *sim2, 1e-7));
    }
}

TEST(TESTNAME, PolarizedFeNiBilayerSpinFlip)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFeNiBilayerSpinFlip());
    for (const std::string polKey : {"PP", "PM", "MP", "MM"}) {
        auto sim1 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerSpinFlip" + polKey, *sim1, 1e-7));
        auto sim2 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerSpinFlip" + polKey + "_Q", *sim2, 1e-7));
    }
}

TEST(TESTNAME, PolarizedFeNiBilayerSpinFlipTanh)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFeNiBilayerSpinFlipTanh());
    for (const std::string polKey : {"PP", "PM", "MP", "MM"}) {
        auto sim1 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerSpinFlipTanh" + polKey, *sim1, 1e-7));
        auto sim2 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerSpinFlipTanh" + polKey + "_Q", *sim2, 1e-7));
    }
}

TEST(TESTNAME, PolarizedFeNiBilayerSpinFlipNC)
{
    std::unique_ptr<const MultiLayer> sample(ExemplarySamples::createFeNiBilayerSpinFlipNC());
    for (const std::string polKey : {"PP", "PM", "MP", "MM"}) {
        auto sim1 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, false);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerSpinFlipNC" + polKey, *sim1, 1e-7));
        auto sim2 = test::makeSimulation::BasicYPolarizedSpecular(*sample, polKey, true);
        EXPECT_TRUE(runTest("PolarizedFeNiBilayerSpinFlipNC" + polKey + "_Q", *sim2, 1e-7));
    }
}

TEST(TESTNAME, Depthprobe)
{
    auto* sample = ExemplarySamples::createHomogeneousMultilayer();
    auto sim = test::makeSimulation::BasicDepthprobe(*sample);
    EXPECT_TRUE(runTest("DepthprobeTest", *sim, 1e-10));
}

#endif // GUI_STD_TEST
#endif // PYTHON_STD_TEST

#endif // BORNAGAIN_TESTS_SUITE_COMMON_TESTSUITE_H
