#include "Base/Const/Units.h"
#include "Base/Py/PyFmt.h"
#include "Param/Distrib/Distributions.h"
#include "Param/Distrib/ParameterDistribution.h"
#include "Sample/Correlations/Profiles1D.h"
#include "Sample/Correlations/Profiles2D.h"
#include "Sample/HardParticle/Bar.h"
#include "Sample/HardParticle/Cone.h"
#include "Sample/HardParticle/CosineRipple.h"
#include "Sample/HardParticle/Cylinder.h"
#include "Sample/HardParticle/EllipsoidalCylinder.h"
#include "Sample/HardParticle/HemiEllipsoid.h"
#include "Sample/HardParticle/HorizontalCylinder.h"
#include "Sample/HardParticle/LongBoxGauss.h"
#include "Sample/HardParticle/LongBoxLorentz.h"
#include "Sample/HardParticle/Polyhedra.h"
#include "Sample/HardParticle/SawtoothRipple.h"
#include "Sample/HardParticle/Sphere.h"
#include "Sample/HardParticle/Spheroid.h"
#include "Sample/HardParticle/TruncatedSphere.h"
#include "Sample/HardParticle/TruncatedSpheroid.h"
#include "Sample/Interface/LayerRoughness.h"
#include "Sample/SoftParticle/FuzzySphere.h"
#include "Sample/SoftParticle/Gauss.h"
#include "Sim/Export/PyFmt2.h"
#include "Sim/Export/PyFmtLimits.h"
#include "Tests/GTestWrapper/google_test.h"

TEST(PythonFormattingTest, ValueTimesUnits)
{
    EXPECT_EQ("2*nm", Py::Fmt::printValue(2.0, "nm"));
    EXPECT_EQ("2*deg", Py::Fmt::printValue(2.0 * Units::deg, "rad"));
    EXPECT_EQ("2.0", Py::Fmt::printValue(2.0, ""));
}

TEST(PythonFormattingTest, RealLimits)
{
    EXPECT_EQ("RealLimits.lowerLimited(1.0)",
              Py::Fmt::printRealLimits(RealLimits::lowerLimited(1.0)));
    EXPECT_EQ("RealLimits.lowerLimited(1*nm)",
              Py::Fmt::printRealLimits(RealLimits::lowerLimited(1.0), "nm"));
    EXPECT_EQ("RealLimits.lowerLimited(1*deg)",
              Py::Fmt::printRealLimits(RealLimits::lowerLimited(1.0 * Units::deg), "rad"));

    EXPECT_EQ("RealLimits.upperLimited(1.0)",
              Py::Fmt::printRealLimits(RealLimits::upperLimited(1.0)));
    EXPECT_EQ("RealLimits.upperLimited(1*nm)",
              Py::Fmt::printRealLimits(RealLimits::upperLimited(1.0), "nm"));
    EXPECT_EQ("RealLimits.upperLimited(1*deg)",
              Py::Fmt::printRealLimits(RealLimits::upperLimited(1.0 * Units::deg), "rad"));

    EXPECT_EQ("RealLimits.limited(1.0, 2.0)",
              Py::Fmt::printRealLimits(RealLimits::limited(1.0, 2.0)));
    EXPECT_EQ("RealLimits.limited(1*nm, 2*nm)",
              Py::Fmt::printRealLimits(RealLimits::limited(1.0, 2.0), "nm"));
    EXPECT_EQ(
        "RealLimits.limited(1*deg, 2*deg)",
        Py::Fmt::printRealLimits(RealLimits::limited(1.0 * Units::deg, 2.0 * Units::deg), "rad"));

    // testing printing of real limits in the context of ParameterDistribution and similar
    EXPECT_EQ(", ba.RealLimits.limited(1*deg, 2*deg)",
              Py::Fmt::printRealLimitsArg(RealLimits::limited(1.0 * Units::deg, 2.0 * Units::deg),
                                          "rad"));

    EXPECT_EQ("", Py::Fmt::printRealLimitsArg(RealLimits::limitless()));
}

TEST(PythonFormattingTest, printDistribution)
{
    EXPECT_EQ(DistributionGate(1.0, 2.0).pythonConstructor(""), "DistributionGate(1.0, 2.0)");

    EXPECT_EQ(DistributionLorentz(1.0, 2.0).pythonConstructor(""), "DistributionLorentz(1.0, 2.0)");

    EXPECT_EQ(DistributionGaussian(1.0, 2.0).pythonConstructor(""),
              "DistributionGaussian(1.0, 2.0)");

    EXPECT_EQ(DistributionGate(1.0, 2.0).pythonConstructor("nm"), "DistributionGate(1*nm, 2*nm)");

    EXPECT_EQ(DistributionGate(1.0 * Units::deg, 2.0 * Units::deg).pythonConstructor("rad"),
              "DistributionGate(1*deg, 2*deg)");

    EXPECT_EQ(DistributionLogNormal(1.0 * Units::deg, 0.01).pythonConstructor("rad"),
              "DistributionLogNormal(1*deg, 0.01)");

    EXPECT_EQ(DistributionCosine(1.0, 2.0).pythonConstructor(""), "DistributionCosine(1.0, 2.0)");

    EXPECT_EQ(DistributionTrapezoid(1.0, 2.0, 3.0, 4.0).pythonConstructor(""),
              "DistributionTrapezoid(1.0, 2.0, 3.0, 4.0)");
}

TEST(PythonFormattingTest, printProfile1D)
{
    EXPECT_EQ(Profile1DCauchy(2.2).pythonConstructor(), "Profile1DCauchy(2.2*nm)");
    EXPECT_EQ(Profile1DGauss(2.2).pythonConstructor(), "Profile1DGauss(2.2*nm)");
    EXPECT_EQ(Profile1DGate(2.2).pythonConstructor(), "Profile1DGate(2.2*nm)");
    EXPECT_EQ(Profile1DTriangle(2.2).pythonConstructor(), "Profile1DTriangle(2.2*nm)");
    EXPECT_EQ(Profile1DCosine(2.2).pythonConstructor(), "Profile1DCosine(2.2*nm)");
    EXPECT_EQ(Profile1DVoigt(2.2, .3).pythonConstructor(), "Profile1DVoigt(2.2*nm, 0.3)");
}

TEST(PythonFormattingTest, printProfile2D)
{
    EXPECT_EQ(Profile2DCauchy(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DCauchy(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DGauss(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DGauss(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DGate(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DGate(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DCone(2.2, 3.3, 4.4 * Units::deg).pythonConstructor(),
              "Profile2DCone(2.2*nm, 3.3*nm, 4.4*deg)");
    EXPECT_EQ(Profile2DVoigt(2.2, 3.3, 4.4 * Units::deg, .5).pythonConstructor(),
              "Profile2DVoigt(2.2*nm, 3.3*nm, 4.4*deg, 0.5)");
}

TEST(PythonFormattingTest, printParameterDistribution)
{
    DistributionGate gate(1.0, 2.0);
    ParameterDistribution dist(ParameterDistribution::BeamWavelength, gate, 5, 2.0);

    // No RealLimits defined
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist, "distr_1"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamWavelength, distr_1, 5, 2.0)");

    // RealLimits defined, units unknown
    ParameterDistribution dist2(ParameterDistribution::BeamWavelength, gate, 5, 2.0,
                                RealLimits::limited(1.0, 2.0));
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist2, "distr_1"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamWavelength, distr_1, "
              "5, 2.0, ba.RealLimits.limited(1.0, 2.0))");

    // RealLimits defined, units explicitly set
    ParameterDistribution dist3(ParameterDistribution::BeamWavelength, gate, 5, 2.0,
                                RealLimits::limited(1.0, 2.0));
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist3, "distr_1", "nm"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamWavelength, distr_1, "
              "5, 2.0, ba.RealLimits.limited(1*nm, 2*nm))");

    // RealLimits defined, checking that method guess radians units correctly
    ParameterDistribution dist4(ParameterDistribution::BeamAzimuthalAngle, gate, 5, 2.0,
                                RealLimits::limited(1.0 * Units::deg, 2.0 * Units::deg));
    EXPECT_EQ(Py::Fmt2::printParameterDistribution(dist4, "distr_1", "rad"),
              "ba.ParameterDistribution(ba.ParameterDistribution.BeamAzimuthalAngle, "
              "distr_1, 5, 2.0, ba.RealLimits.limited(1*deg, 2*deg))");
}

TEST(PythonFormattingTest, printLayerRoughness)
{
    EXPECT_EQ(LayerRoughness(1.1, 0.2, 3.3).pythonConstructor(),
              "LayerRoughness(1.1, 0.2, 3.3*nm)");
}

TEST(PythonFormattingTest, printFormFactors)
{
    EXPECT_EQ(Pyramid2(12.0, 14.0, 5.0, 55 * Units::deg).pythonConstructor(),
              "Pyramid2(12*nm, 14*nm, 5*nm, 55*deg)");

    EXPECT_EQ(GaussSphere(12.0).pythonConstructor(), "GaussSphere(12*nm)");

    EXPECT_EQ(HemiEllipsoid(6, 7, 5).pythonConstructor(), "HemiEllipsoid(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(Box(6, 7, 5).pythonConstructor(), "Box(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(Cone(6, 5, 55 * Units::deg).pythonConstructor(), "Cone(6*nm, 5*nm, 55*deg)");

    EXPECT_EQ(Pyramid6(6, 5, 55 * Units::deg).pythonConstructor(), "Pyramid6(6*nm, 5*nm, 55*deg)");

    EXPECT_EQ(Bipyramid4(10, 4, 0.7, 55 * Units::deg).pythonConstructor(),
              "Bipyramid4(10*nm, 4*nm, 0.7, 55*deg)");

    EXPECT_EQ(Cylinder(3, 5).pythonConstructor(), "Cylinder(3*nm, 5*nm)");

    // should we expect "HorizontalCylinder(3*nm, 5*nm)" from pythonConstructor() ?
    //    EXPECT_EQ(HorizontalCylinder(3, 5).pythonConstructor(),
    //              "HorizontalCylinder(3*nm, 5*nm, -3*nm, 3*nm)");

    EXPECT_EQ(HorizontalCylinder(3, 5, -2, 2).pythonConstructor(),
              "HorizontalCylinder(3*nm, 5*nm, -2*nm, 2*nm)");

    EXPECT_EQ(Dodecahedron(3).pythonConstructor(), "Dodecahedron(3*nm)");

    EXPECT_EQ(EllipsoidalCylinder(3, 4, 5).pythonConstructor(),
              "EllipsoidalCylinder(3*nm, 4*nm, 5*nm)");

    EXPECT_EQ(CantellatedCube(10, 2).pythonConstructor(), "CantellatedCube(10*nm, 2*nm)");

    EXPECT_EQ(Sphere(5).pythonConstructor(), "Sphere(5*nm)");

    EXPECT_EQ(Spheroid(3, 5).pythonConstructor(), "Spheroid(3*nm, 5*nm)");

    EXPECT_EQ(Icosahedron(7).pythonConstructor(), "Icosahedron(7*nm)");

    EXPECT_EQ(LongBoxGauss(6, 7, 5).pythonConstructor(), "LongBoxGauss(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(LongBoxLorentz(6, 7, 5).pythonConstructor(), "LongBoxLorentz(6*nm, 7*nm, 5*nm)");

    EXPECT_EQ(PlatonicOctahedron(7).pythonConstructor(), "PlatonicOctahedron(7*nm)");

    EXPECT_EQ(PlatonicTetrahedron(7).pythonConstructor(), "PlatonicTetrahedron(7*nm)");

    EXPECT_EQ(Prism3(6, 4).pythonConstructor(), "Prism3(6*nm, 4*nm)");

    EXPECT_EQ(Prism6(3, 4).pythonConstructor(), "Prism6(3*nm, 4*nm)");

    EXPECT_EQ(Pyramid4(10, 4, 60 * Units::deg).pythonConstructor(),
              "Pyramid4(10*nm, 4*nm, 60*deg)");

    EXPECT_EQ(TruncatedSphere(5, 3, 0).pythonConstructor(), "TruncatedSphere(5*nm, 3*nm, 0*nm)");

    EXPECT_EQ(TruncatedSpheroid(3, 5, 1.5, 0).pythonConstructor(),
              "TruncatedSpheroid(3*nm, 5*nm, 1.5, 0*nm)");

    EXPECT_EQ(Pyramid3(16, 4, 60 * Units::deg).pythonConstructor(),
              "Pyramid3(16*nm, 4*nm, 60*deg)");

    EXPECT_EQ(CosineRippleBox(100, 20, 4).pythonConstructor(),
              "CosineRippleBox(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(CosineRippleGauss(100, 20, 4).pythonConstructor(),
              "CosineRippleGauss(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(CosineRippleLorentz(100, 20, 4).pythonConstructor(),
              "CosineRippleLorentz(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(TruncatedCube(15, 6).pythonConstructor(), "TruncatedCube(15*nm, 6*nm)");

    EXPECT_EQ(SawtoothRippleBox(100, 20, 4, 0.3).pythonConstructor(),
              "SawtoothRippleBox(100*nm, 20*nm, 4*nm, 0.3*nm)");

    EXPECT_EQ(SawtoothRippleGauss(100, 20, 4, 0.3).pythonConstructor(),
              "SawtoothRippleGauss(100*nm, 20*nm, 4*nm, 0.3*nm)");

    EXPECT_EQ(SawtoothRippleLorentz(100, 20, 4, 0.3).pythonConstructor(),
              "SawtoothRippleLorentz(100*nm, 20*nm, 4*nm, 0.3*nm)");

    EXPECT_EQ(FuzzySphere(100, 20).pythonConstructor(), "FuzzySphere(100*nm, 20*nm)");

    EXPECT_EQ(BarGauss(100, 20, 4).pythonConstructor(), "BarGauss(100*nm, 20*nm, 4*nm)");

    EXPECT_EQ(BarLorentz(100, 20, 4).pythonConstructor(), "BarLorentz(100*nm, 20*nm, 4*nm)");
}
